/*
 * fanxdell
 * Userspace fancontrol for use while installing Debian and as a system daemon
 * after it is installed. Also see infodell, shows fan settings and other data.
 * Copyright © 2014 - 2016 C_Xperimental
 * Not associated in any way with the company Dell.
 * Not associated in any way with i8kutils or dellfand, although several
 * functions have been adapted from their source code. The interface with
 * the BIOS is very similar but the rest is original.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * <www.cxperimental.weebly.com>
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/io.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <syslog.h>
#include <fcntl.h>
#include <signal.h>
#include <ctype.h>

/* Uncomment line below for debug, try tail -f /var/log/syslog for daemon. */
/*#define DEBUG 1*/

/* Uncomment line below to disable make/model safety checks, then compile. */
/*#define UNSAFE_MODE 1*/

#define USER_ARG_FILE "/etc/default/fanxdell"
#define PID_FILE_NAME "fanxdell_8sl7m21wdym4.pid" /* Modifiable, malloc'ed */
#define DMI_MAKE "/sys/class/dmi/id/sys_vendor"
#define DMI_MODEL "/sys/class/dmi/id/product_name"
#define PROGRAM_VERSION 3

/* Signal interrupt handling */
#define X_FINITE_FAILURE_RETRY(expression) \
  (__extension__                                                              \
    ({ long int result;                                                       \
       long int index = 0;                                                    \
       do result = (long int) (expression);                                   \
       while (result == -1L && errno == EINTR && (index++) < 1000);          \
       result; }))

typedef struct
{
  unsigned int eax;
  unsigned int ebx __attribute__ ((packed));
  unsigned int ecx __attribute__ ((packed));
  unsigned int edx __attribute__ ((packed));
  unsigned int esi __attribute__ ((packed));
  unsigned int edi __attribute__ ((packed));
} SMMRegisters;

struct fan_data
{
  int what_fan; /* 0 left, 1 right, 2 both */
  int poll;
  int t01;
  int t02;
  int h01;
  int h02;
  int t11;
  int t12;
  int h11;
  int h12;
};

void sigf (char *sig, int err);
void *xmalloc (size_t size);
void *xrealloc (void *ptr, size_t size);
void signal_setup (void);
void signal_unblock (void);
void kill_option (void) __attribute__ ((noreturn));
int lock_n_kill (char *killpid, int killopt);
int dell_smm (SMMRegisters *regs);
int get_cpu_temp (void);
int get_fan_status (int fan);
int set_fan (int fan, int speed);
int get_dell_sig (int cmd, unsigned int *eax, unsigned int *edx);
#ifndef UNSAFE_MODE
void dmi_check (void);
#endif
int get_pid_dir (void);
void set_lock (int how);
void pid_file (void);
void data_write (void);
void fan_daemon (int fork) __attribute__ ((noreturn, hot));
void typo (void) __attribute__ ((noreturn));
int get_args (char *in_str);
void get_ioperm (void);
int main (int argc, char **argv);

/* Global variables for signal handling */
volatile sig_atomic_t usr1 = 0, usr2 = 0, fin = 0, newpid = 0, sigsave = 0;

/* More global variables */
char **argB = NULL;
char *pid_dir;
int fd;
struct fan_data dat =
{
  .what_fan = 0,
  .poll = 10,
  .t01 = 35, /* fan0 if using 2 fans, 'the' fan if using only one */
  .t02 = 62, /* t = temperature threshold */
  .h01 = 5,  /* h = hysteresis buffer */
  .h02 = 8,
  .t11 = 47, /* fan1 if using 2 fans */
  .t12 = 62,
  .h11 = 7,
  .h12 = 8
};
char usage[] =
"fanxdell --help / -?  : prints this help.\n\n\
fanxdell --fan / -f  [-/0/1/2] [-/0/1/2]  : alters fan speed\n\n\
fanxdell --daemon / -d fan poll t01 t02 h01 h02 t11 t12 h11 h12\n\
Starts a fan control program, or replaces a running fanxdell.\n\
fan is the only required argument. Select 0, 1, or for twin fans use 2.\n\
t01 = turn on fan 0 at low speed at this temperature.\n\
h12 = hysteresis for high speed threshold (t12) for fan 1, and so on.\n\
Any preceding arguments must be given to set later values.\n\
Default settings for single fan:\n\
fanxdell --daemon [0/1] 10 35 62 5 8\n\
Default settings for twin fan:\n\
fanxdell --daemon 2 10 35 62 5 8 47 62 7 8\n\
Unset values stay at defaults.\n\
Temperature thresholds and hysteresis values must be between 0 and 127 c.\n\
Polling frequency must be between 1 and 30 seconds.\n\
Simplest use: fanxdell -d [0/1/2]\n\n\
fanxdell --auto / -a poll t01 t02 h01 h02 t11 t12 h11 h12\n\
Identical to --daemon option but automatically detects fans.\n\
Simplest use:  fanxdell -a\n\n\
fanxdell --kill / -k  : Stops any running fanxdell daemon.\n\n\
fanxdell --version / -V  : prints program version.\n\n\
fanxdell --readargs [--foreground] : Intended for init files.\n\
Reads /etc/default/fanxdell or defaults to -a, optionally don't fork.\n\
\n";

/* Signal handler functions */
void
usr1_handler (int signum, siginfo_t *info, void *ptr __attribute__ ((unused)))
{
  usr1++;
  sigsave = signum;
  newpid = info->si_pid;
}

void
usr2_handler (int signum __attribute__ ((unused)))
{
  usr2++;
}

void
fin_handler (int signum)
{
  fin++;
  sigsave = signum;
}

void
sigf (char *sig, int err)
{
  errno = err;
  fprintf (stderr, "%s, sigaction failure : %m\n", sig);
  exit (EXIT_FAILURE);
}

void *
xmalloc (size_t size)
{
  errno = 0;
  void *value = malloc (size);
  if (value == 0)
    {
      perror ("virtual memory exhausted");
      exit (EXIT_FAILURE);
    }
  return value;
}

void *
xrealloc (void *ptr, size_t size)
{
  errno = 0;
  void *value = realloc (ptr, size);
  if (value == 0)
    {
      perror ("virtual memory exhausted");
      exit (EXIT_FAILURE);
    }
  return value;
}

/* Setup signal handler functions for various signals */
void
signal_setup (void)
{
  struct sigaction newaction, oldaction;
  newaction.sa_sigaction = usr1_handler;
  sigemptyset (&newaction.sa_mask);
  newaction.sa_flags = SA_RESTART | SA_SIGINFO; /* SA_SIGINFO for 3 args */
  errno = 0;
  if (sigaction (SIGUSR1, &newaction, NULL)) sigf ("SIGUSR1", errno);
  newaction.sa_handler = usr2_handler;
  newaction.sa_flags = SA_RESTART;
  errno = 0;
  if (sigaction (SIGUSR2, &newaction, NULL)) sigf ("SIGUSR2", errno);
  newaction.sa_handler = fin_handler;
  oldaction.sa_handler = NULL;
  sigemptyset (&oldaction.sa_mask);
  errno = 0;
  if (sigaction (SIGTERM, NULL, &oldaction)) sigf ("SIGTERM query", errno);
  if (oldaction.sa_handler != SIG_IGN)
    {
      errno = 0;
      if (sigaction (SIGTERM, &newaction, NULL)) sigf ("SIGTERM", errno);
    }
  oldaction.sa_handler = NULL;
  errno = 0;
  if (sigaction (SIGINT, NULL, &oldaction)) sigf ("SIGINT query", errno);
  if (oldaction.sa_handler != SIG_IGN)
    {
      errno = 0;
      if (sigaction (SIGINT, &newaction, NULL)) sigf ("SIGINT", errno);
    }
  oldaction.sa_handler = NULL;
  errno = 0;
  if (sigaction (SIGQUIT, NULL, &oldaction)) sigf ("SIGQUIT query", errno);
  if (oldaction.sa_handler != SIG_IGN)
    {
      errno = 0;
      if (sigaction (SIGQUIT, &newaction, NULL)) sigf ("SIGQUIT", errno);
    }
  newaction.sa_handler = SIG_IGN; /* SIGHUP - reload conf - no need, ignore */
  oldaction.sa_handler = NULL;
  errno = 0;
  if (sigaction (SIGHUP, NULL, &oldaction)) sigf ("SIGHUP query", errno);
  if (oldaction.sa_handler != SIG_IGN)
    {
      errno = 0;
      if (sigaction (SIGHUP, &newaction, NULL)) sigf ("SIGHUP", errno);
    }
}/* Fatal sigs not sent to orphaned process groups SIGTSTP SIGTTIN SIGTTOU */

/* Helper function used during cleanup when daemon exits */
void
signal_unblock (void)
{
 struct sigaction newaction;
  newaction.sa_handler = SIG_DFL;
  sigemptyset (&newaction.sa_mask);
  errno = 0;
  if (sigaction (sigsave, &newaction, NULL))
    {
      syslog (LOG_ERR, "sigaction failure at cleanup : %m");
      exit (EXIT_FAILURE);
    }
}

/* Manages --kill option */
void
kill_option (void) /* Find pidfile(s) first */
{
  int flag = 0;
  struct stat test_stat;
  /* Allocate memory: pid name length + directory (11) + null (1) */
  /* If the directory names are changed this may need changing. */
  signal_setup ();
  errno = 0;
  pid_dir = (char *) xmalloc ((strlen (PID_FILE_NAME)) + (size_t) 12);
  if ((!(stat ("/run/local", &test_stat))) &&
      (S_ISDIR (test_stat.st_mode))) /* Is it a directory? */
    {
      strcpy (pid_dir, "/run/local/");
      strcat (pid_dir, PID_FILE_NAME);
      if ((!(stat (pid_dir, &test_stat))) && (S_ISREG (test_stat.st_mode)))
        {
          /* Its a pidfile */
          if (lock_n_kill (pid_dir, 1))
            fprintf (stderr, "Error processing %s\n", pid_dir);
          else flag++;
        }
    }
  if ((!(stat ("/tmp/local", &test_stat))) &&
      (S_ISDIR (test_stat.st_mode)))
    {
      strcpy (pid_dir, "/tmp/local/");
      strcat (pid_dir, PID_FILE_NAME);
      if ((!(stat (pid_dir, &test_stat))) && (S_ISREG (test_stat.st_mode)))
        {
          /* Its a pidfile */
          if (lock_n_kill (pid_dir, 1))
            fprintf (stderr, "Error processing %s\n", pid_dir);
          else flag++;
        }
    }
  if (!flag)
    printf ("No pidfile detected, no fanxdell daemon was running.\n");
  exit (EXIT_SUCCESS);
}

/* Helper function for kill_option */
int
lock_n_kill (char *killpid, int killopt)
{
  int i;
  struct flock lk_dat;
  /* Open and get a file descriptor */
  errno = 0;
  fd = X_FINITE_FAILURE_RETRY (open (killpid, O_RDWR | O_NONBLOCK,
                               S_IWUSR | S_IRUSR));
  if (fd < 0)
    {
      if (killopt)
        {
          perror ("Pidfile open failed");
          fprintf (stderr, "Could not get a file descriptor for %s\n",
                   killpid);
        }
      return (1); /* Because alternate directories can be checked */
    }
  for (i = 3; i >= 1; i--)
    {
      lk_dat.l_type = F_WRLCK;
      lk_dat.l_whence = SEEK_SET;
      lk_dat.l_start = (off_t) 0;
      lk_dat.l_len = (off_t) 1;
      lk_dat.l_pid = (pid_t) 0;
      errno = 0;
      if (fcntl (fd, F_GETLK, (&lk_dat)))
        {
          if (killopt)
            {
              perror ("fcntl F_GETLK failed");
              fprintf (stderr, "Could not get lock status for %s\n", killpid);
            }
          if (i <= 1) return (1); /* continue and return on last loop only */
          else continue;
        }
      if (lk_dat.l_type == F_UNLCK)
        {
          /* No active lock, stale pidfile */
          if (killopt)
            {
              if (i == 3) printf ("Unlocked pidfile, checking for daemon -\n");
              printf ("Seconds remaining : %d\n", i);
            }
          /* Loop in case of a daemon being started during this check */
          if (i <= 1)
            {
              errno = 0;
              if (unlink (killpid))
                {
                  if (killopt)
                    {
                      perror ("First unlink failed in lock_n_kill");
                      fprintf (stderr, "Couldn't delete %s\n", killpid);
                    }
                  return (1);
                }
              if (killopt) printf ("Deleted unlocked pidfile : %s\n", killpid);
              return (0);
             }
        }
      if ((lk_dat.l_type == F_WRLCK) && (lk_dat.l_pid > 0))
        {
          break;
        }
      sleep (1);
    } /* end of for loop */

  if (lk_dat.l_type != F_WRLCK)
    {
      if (killopt) (fprintf (stderr, "File %s has an unexpected lock status\n",
                    killpid));
      return (1);
    }
  if (killopt) printf ("Active lock on %s\n", killpid);
  /* File locked, kill daemon */
  errno = 0;
  if (kill (lk_dat.l_pid, SIGUSR1))
    {
      if (killopt) perror ("kill failed to send signal");
      return (1);
    }
  for (i = 5; i >= 1; i--)
    {
      if (usr2 > 0)
        {
          /* Previous daemon has released lock, delete pidfile */
          if (killopt) printf ("Daemon returned SIGUSR2, unlinking %s\n",
                               killpid);
          errno = 0;
          if (unlink (killpid))
            {
              if (killopt)
                {
                  perror ("Second unlink failed in lock_n_kill");
                  fprintf (stderr, "Couldn't delete %s\n", killpid);
                }
              return (1);
            }
          return (0);
        }
      if (killopt)
        {
          if (i == 5) printf ("Waiting for SIGUSR2\n");
          printf ("Seconds to timeout = %d\n", i);
        }
      sleep (1);
      /* Otherwise keep looping until SIGUSR2 is received */
    }
  /* Unknown proc locks pidfile, fanxdell malfunction, CPU o/load? */
  if (killopt) printf ("%s locked by unresponsive process\npid no. = %d\n",
          killpid, lk_dat.l_pid);
  errno = 0;
  if (kill (lk_dat.l_pid, SIGKILL))
    {
      if (killopt) perror ("kill failed to send SIGKILL");
      return (1);
    }
  if (killopt)
    {
      printf ("SIGKILL sent to %d\n", lk_dat.l_pid);
      printf ("Unlinking %s\n", killpid);
    }
  errno = 0;
  if (unlink (killpid))
    {
      if (killopt)
        {
          perror ("Third unlink failed in lock_n_kill");
          fprintf (stderr, "Couldn't delete %s\n", killpid);
        }
      return (1);
    }
  return (0);
}

/* Write 1 byte to 0xb2, 1 byte to 0x84, put CPU in system management mode */
/* 0xb2 = Dell command register, 0x84 = ACPI SMB_DATA[0] Data register zero */
int
dell_smm (SMMRegisters *regs)
{
  int rc;
  unsigned int eax = regs->eax;
  asm volatile ("pushl %%eax\n\t" \
      "movl 0(%%eax),%%edx\n\t" \
      "push %%edx\n\t" \
      "movl 4(%%eax),%%ebx\n\t" \
      "movl 8(%%eax),%%ecx\n\t" \
      "movl 12(%%eax),%%edx\n\t" \
      "movl 16(%%eax),%%esi\n\t" \
      "movl 20(%%eax),%%edi\n\t" \
      "popl %%eax\n\t" \
      "out %%al,$0xb2\n\t" \
      "out %%al,$0x84\n\t" \
      "xchgl %%eax,(%%esp)\n\t"
      "movl %%ebx,4(%%eax)\n\t" \
      "movl %%ecx,8(%%eax)\n\t" \
      "movl %%edx,12(%%eax)\n\t" \
      "movl %%esi,16(%%eax)\n\t" \
      "movl %%edi,20(%%eax)\n\t" \
      "popl %%edx\n\t" \
      "movl %%edx,0(%%eax)\n\t" \
      "lahf\n\t" \
      "shrl $8,%%eax\n\t" \
      "andl $1,%%eax\n" \
      : "=a" (rc)
      : "a" (regs)
      : "%ebx", "%ecx", "%edx", "%esi", "%edi", "memory");
  if ((regs->eax & 0xffff) == 0xffff) return (-EINVAL);
  /* The above case MUST be handled separately for auto fan detection */
  if ((rc != 0) || (regs->eax == eax)) return (-EAGAIN);
  return (0);
}  /* i386: EINVAL = 22 EAGAIN = 11 (- = no confusion with valid result) */

/* Read the cpu temperature. */
int
get_cpu_temp (void)
{
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};
  int temp, err;
  /* Set to 0 on initialisation only. */
  static int prev = 0;
  regs.eax = 0x10a3;
  err = dell_smm (&regs);
  if (err) return (err);
  temp = regs.eax & 0xff;
  /* Sometimes the temperature sensor returns 0x99, which is out of range. */
  /* In this case we return (once) the previous cached value. */
  if (temp > 127)
    {
      temp = prev;
      prev = 127;
    }
  else
    {
      prev = temp;
    }
  return (temp);
}

/* Current fan setting, off, low, or high: ie. 0, 1, or 2 */
int
get_fan_status (int fan)
{
  int err;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};
  regs.eax = 0x00a3;
  regs.ebx = fan & 0xff;
  err = dell_smm (&regs);
  if (err) return (err);
  return (regs.eax & 0xff);
}

/* Alters fan speed, fan = 0 or 1, speed = 0, 1 or 2 */
int
set_fan (int fan, int speed)
{
  int err;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};

  /* set the register values with command and argument */
  regs.eax = 0x01a3;
  regs.ebx = (fan & 0xff) | (speed << 8);
  /* call the assembly code function */
  err = dell_smm (&regs);
  return (err);
}

/* Read DIAG DELL signatures from BIOS */
int
get_dell_sig (int cmd, unsigned int *eax, unsigned int *edx)
{
  int err;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};

  regs.eax = cmd;
  err = dell_smm (&regs);
  if (err) return (err);
  *eax = regs.eax;
  *edx = regs.edx;
  return (0);
}

/* Read sysfs kernel files to check this is a compatible machine */
#ifndef UNSAFE_MODE
void
dmi_check (void)
{
  char *str1 = NULL;
  ssize_t get_ret = (ssize_t) 0;
  size_t n_bytes = (size_t) 0;
  FILE *stream1;
  int dell = 0;
  int livp = 0;

  errno = 0;
  stream1 = fopen (DMI_MAKE, "r");
  if (stream1 == NULL)
    {
      perror ("fopen");
      fprintf (stderr, "Could not open %s\n", DMI_MAKE);
      exit (EXIT_FAILURE);
    }
  do
    {
      get_ret = getline (&str1, &n_bytes, stream1);
      /* Check make is Dell, strstr returns null for no match */
      if (strstr (str1, "Dell"))
        {
#ifdef DEBUG
          fprintf (stderr, "Well, it's a Dell\n");
#endif
          dell = 1;
          break;
        }
    }
  while (get_ret != -1);
  /* Runs through loop twice in practice, not changing str1 2nd time (eof) */
  if (str1 != NULL)
    {
      free (str1);
      str1 = NULL;
    }
  n_bytes = (size_t) 0; /* str1, n_bytes reset for next getline */
  fclose (stream1);
  errno = 0;
  stream1 = fopen (DMI_MODEL, "r");
  if (stream1 == NULL)
    {
      perror ("fopen");
      fprintf (stderr, "Could not open %s\n", DMI_MODEL);
      exit (EXIT_FAILURE);
    }
  do
    {
      get_ret = getline (&str1, &n_bytes, stream1);
      /* Check for supported model name */
      if ((strstr (str1, "Latitude")) || (strstr (str1, "Inspiron"))
          || (strstr (str1, "Precision")) || (strstr (str1, "MM061"))
          || (strstr (str1, "MP061")) || (strstr (str1, "Vostro")))
        {
#ifdef DEBUG
          fprintf(stderr, "It's a Latitude, Inspiron, Vostro or Precision\n");
#endif
          livp = 1;
          break;
        }
    }
  while (get_ret != -1);
  fclose (stream1);
  if (str1 != NULL) /* beware the unlikely unexpected seg fault */
    {
      free (str1); /* free the bytes */
    }
  if (dell == 0)
    {
      fprintf (stderr, "Cannot find Dell in %s\n", DMI_MAKE);
    }
  if (livp == 0)
    {
      fprintf (stderr,
               "Latitude, Inspiron, Precision, Vostro, MM061, MP061\
                not found in %s\n", DMI_MODEL);
    }
  if ((dell == 0) || (livp == 0))
    {
      fprintf (stderr,
               "Your computer may be incompatible, try further research\n");
      exit (EXIT_FAILURE);
    }
}
#endif

/* Location for pidfile - /run/local, or /tmp/local. */
int
get_pid_dir (void)
{
  struct stat test_stat;

  if ((stat ("/run", &test_stat)) == 0)
    {
      if ((!(stat ("/run/local", &test_stat))) &&
          (S_ISDIR (test_stat.st_mode))) /* is it a directory? */
        {
          /* Use "/run/local/" */
          return (1);
        }
      if (!(mkdir ("/run/local", S_IRWXU | S_IRGRP | S_IXGRP
          | S_IROTH | S_IXOTH)))
          /* fails if /run wasn't a dir */
        {
          /* Use "/run/local/" */
          return (1);
        }
      syslog (LOG_ERR, "mkdir: failed to create /run/local");
    }
  if ((stat ("/tmp", &test_stat)) == 0)
    {
      if ((!(stat ("/tmp/local", &test_stat))) &&
          (S_ISDIR (test_stat.st_mode)))
        {
          /* Use "/tmp/local/" */
          return (0);
        }
      if (!(mkdir ("/tmp/local", S_IRWXU | S_IRGRP | S_IXGRP
          | S_IROTH | S_IXOTH)))
        {
          /* Use "/tmp/local/" */
          return (0);
        }
      syslog (LOG_ERR, "mkdir: failed to create /tmp/local");
    }
  if (mkdir ("/run", S_IRWXU | S_IRGRP | S_IXGRP
      | S_IROTH | S_IXOTH))
    {
      syslog (LOG_ERR, "mkdir: failed to create /run, quitting.");
      exit (EXIT_FAILURE);
    }
  if (mkdir ("/run/local", S_IRWXU | S_IRGRP | S_IXGRP
      | S_IROTH | S_IXOTH))
    {
      syslog (LOG_ERR, "mkdir: failed to create /run/local, quitting.");
      exit (EXIT_FAILURE);
    }
  /* Use "/run/local/" */
  return (1);
}

/* how = 1 for write lock, how = 2 to unlock */
void
set_lock (int how)
{
  struct flock lk_dat;
  /* fcntl with these cmds returns 0, or -1 with errno set. */
  if (how == 1) lk_dat.l_type = F_WRLCK;
  if (how == 2) lk_dat.l_type = F_UNLCK;
  lk_dat.l_whence = SEEK_SET;
  lk_dat.l_start = (off_t) 0;
  lk_dat.l_len = (off_t) 1;
  errno = 0;
  if (fcntl (fd, F_SETLK, (&lk_dat)))
    {
      syslog (LOG_ERR, "fcntl F_SETLK failed. Error = %m");
      if (how == 1) exit (EXIT_FAILURE);
      /* For unlock, the fd will get closed next anyway. */
    }
}

/* handle fanxdell pidfile in other directory (eg /run created later) */
void
other_pid_dir (int dir)
{
  struct stat test_stat;
  char *alt_dir;

  alt_dir = (char *) malloc ((strlen (PID_FILE_NAME)) + 12);
  if (alt_dir == 0)
    {
      syslog (LOG_ERR, "malloc: Virtual memory exhausted");
      exit (EXIT_FAILURE);
    }
  if (dir) strcpy (alt_dir, "/tmp/local/");
  else strcpy (alt_dir, "/run/local");
  strcat (alt_dir, PID_FILE_NAME);
  if (stat (alt_dir, &test_stat))
    {
      free (alt_dir);
      return; /* Not there, end function */
    }
  if (!(S_ISREG (test_stat.st_mode)))
    {
      free (alt_dir);
      return; /* Not a file, end function */
    }
  if (lock_n_kill (alt_dir, 0)) /* Get rid of pid */
    {
      syslog (LOG_ERR, "lock_n_kill failed in other_pid_dir");
      syslog (LOG_ERR, "Error processing %s", alt_dir);
      exit (EXIT_FAILURE);
    }
  syslog (LOG_ERR, "Unexpectedly processed %s", alt_dir);
  free (alt_dir);
}

/* Create and lock pidfile or get pid of existing fan daemon */
void
pid_file (void)
{
  int count = 0;
  struct flock lk_dat;

  /* Allocate memory: pid name length + directory (11) + null (1) */
  /* If the directories are changed then this may need changing too. */
  pid_dir = (char *) malloc ((strlen (PID_FILE_NAME)) + (size_t) 12);
  if (pid_dir == 0)
    {
      syslog (LOG_ERR, "malloc: Virtual memory exhausted");
      exit (EXIT_FAILURE);
    }
  if (get_pid_dir ())
    {
      strcpy (pid_dir, "/run/local/");
      other_pid_dir (1);
    }
  else
    {
      strcpy (pid_dir, "/tmp/local/");
      other_pid_dir (0);
    }
  strcat (pid_dir, PID_FILE_NAME);
  errno = 0;
  fd = X_FINITE_FAILURE_RETRY (open (pid_dir, O_RDWR | O_CREAT | O_EXCL |
                                O_NONBLOCK, S_IWUSR | S_IRUSR));
  if (fd < 0)
    {
      /* file not opened: an error, or it exists */
      if ((errno == EEXIST) || (errno == EWOULDBLOCK))
        {
          /* probably pidfile already exists */
          errno = 0;
          /* Last open failed, so get a valid file descriptor */
          fd = X_FINITE_FAILURE_RETRY (open (pid_dir, O_RDWR | O_NONBLOCK,
                                        S_IWUSR | S_IRUSR));
          if (fd < 0)
            {
              syslog (LOG_ERR, "Second pidfile open failed. Error = %m");
              exit (EXIT_FAILURE);
            }
          lk_dat.l_type = F_WRLCK;
          lk_dat.l_whence = SEEK_SET;
          lk_dat.l_start = (off_t) 0;
          lk_dat.l_len = (off_t) 1;
          lk_dat.l_pid = (pid_t) 0;
          errno = 0;
          if (fcntl (fd, F_GETLK, (&lk_dat)))
            {
              syslog (LOG_ERR, "fcntl F_GETLK failed. Error = %m");
              exit (EXIT_FAILURE);
            }
          if (lk_dat.l_type == F_UNLCK)
            {
              /* No active lock, stale pidfile, take over and lock it */
#ifdef DEBUG
              syslog (LOG_ERR, "Stale pidfile found, fd = %d", fd);
#endif
              set_lock (1);
              return;
            }
          if ((lk_dat.l_type == F_WRLCK) && (lk_dat.l_pid > 0))
            {
              /* File locked, replace previous daemon */
              errno = 0;
              if (kill (lk_dat.l_pid, SIGUSR1))
                {
                  syslog (LOG_ERR, "kill failed in pid_file. Error = %m");
                  exit (EXIT_FAILURE);
                }
              while (count < 60)
                {
                  count++;
                  sleep (1);
                  if (usr2 > 0)
                    {
                      /* Previous daemon has released pidfile lock */
                      set_lock (1);
#ifdef DEBUG
                      syslog (LOG_ERR, "Daemon returned SIGUSR2, fd = %d", fd);
#endif
                      return;
                    }
                  /* Loop until SIGUSR2 is received or timeout */
                }
              /* Multiple daemons started, unknown process locked pidfile? */
              syslog (LOG_ERR, "Pidfile locked, no response, timed out.");
              exit (EXIT_FAILURE);
            }
          /* Not unlocked, not F_WRLCK, something has a F_RDLCK? */
          syslog (LOG_ERR, "Failure: pidfile locked by another process?");
          exit (EXIT_FAILURE);
        }
      /* errno says it is a definite error */
      syslog (LOG_ERR, "open failed in pid_file. Error = %m");
      exit (EXIT_FAILURE);
    }
  /* open succeeded, set lock to prove this process is active */
  set_lock (1);
#ifdef DEBUG
  syslog (LOG_ERR, "No pidfile, new start, fd = %d", fd);
#endif
  return;
}

/* Write settings data to pidfile for infodell */
void
data_write ()
{
  int tmp = 0;
  off_t f_pos; /* return value from lseek sets this to position or -1 */
  ssize_t byte_no = (ssize_t) 0;
  ssize_t byte_total = (ssize_t) 0;
  char *databuf;

  /* set file position  */
  while (tmp < 5000)
    {
      tmp++; /* On last while tmp = 4999, set to 5000 here */
      errno = 0;
      f_pos = lseek (fd, 0, SEEK_SET);
      if ((f_pos == -1) && (tmp == 5000))
        {
          syslog (LOG_ERR, "lseek failure : %m");
          return;
        }
      if ((f_pos != 0) && (tmp == 5000))
        {
          syslog (LOG_ERR, "lseek problem, f_pos = %d", (int) f_pos);
          return;
        }
      if (f_pos == 0) break;
    }
  /* allocate and initialise buffer for write */
  errno = 0;
  databuf = (char *) malloc (sizeof (struct fan_data));
  if (databuf == NULL)
    {
      syslog (LOG_ERR, "Virtual memory exhausted for databuf : %m");
      exit (EXIT_FAILURE);
    }
  memcpy ((void *) databuf, (const void *) &dat, (sizeof (struct fan_data)));
  tmp = 0;
  /* write struct fan_data dat to file */
  while (tmp < 5000)
    {
      tmp++;
      errno = 0;
      byte_no = write (fd, &databuf[byte_total],
                       ((sizeof (struct fan_data)) - byte_total));
      if ((tmp == 5000) && (byte_no == -1))
        {
          syslog (LOG_ERR, "Error writing databuf : %m");
        }
      if (byte_no > 0) byte_total += byte_no;
#ifdef DEBUG
      if ((tmp < 10) || (tmp > 4995))
        {
          syslog (LOG_ERR, "dat: byte_no = %d, tmp = %d, fd = %d,"
                  " byte_total = %d", byte_no, tmp, fd, byte_total);
        }
#endif
      if (byte_total == (sizeof (struct fan_data))) break;
      if (tmp == 5000)
        {
          if (byte_total == 0)
            {
              syslog (LOG_ERR, "Data for databuf could not be written");
              free (databuf);
              return;
            }
          if (byte_total > 0)
            {
              syslog (LOG_ERR, "Data write for databuf not completed");
              free (databuf);
              return;
            }
        }
    }
  free (databuf);
  errno = 0;
  if (fsync (fd)) syslog (LOG_ERR, "fsync failure : %m");
  return;
}


/* --daemon and --auto option */
void
fan_daemon (int fork_flag)
{
  pid_t pid, sid;
  int tmp = 0, cpu_t = 0, f_stat0 = 0, f_stat1 = 0;

#ifdef DEBUG
  fprintf (stderr, "Preparing to fork, pid file name = %s, fork_flag = %d\n",
           PID_FILE_NAME, fork_flag);
#endif
  signal_setup ();
  if (fork_flag) /* no fork for runit & systemd Type=simple */
    {
      /* fork and write a pid file to /(run or tmp)/local/ */
      errno = 0;
      pid = fork ();
      if (pid < 0)
        {
          perror ("Fork in fan daemon failed");
          exit (EXIT_FAILURE);
        }
      if (pid > 0) exit (EXIT_SUCCESS);
      umask (0); /* set file permission mask to 0 */
      /* open syslog and set values, /dev/log is opened by syslog if needed */
    }
  openlog ("fanxdell", LOG_PID, LOG_DAEMON);
  if (fork_flag)
    {
      errno = 0;
      sid = setsid();
      if (sid < 0)
        {
          syslog (LOG_ERR, "setsid failed. Error = %m");
          exit (EXIT_FAILURE);
        }
      /* 2nd fork PID != SID, not session leader, won't get controlling tty */
      errno = 0; /* systemd handles 1 fork if Type=forking & no PID in 2016*/
      pid = fork (); /* but it can start with 2 forks here if Type=oneshot */
      if (pid < 0)
        {
          syslog (LOG_ERR, "Second fork failed. Error = %m");
          exit (EXIT_FAILURE);
        }
      if (pid > 0) exit (EXIT_SUCCESS);
      errno = 0;
      if ((chdir ("/")) < 0)
        {
          syslog (LOG_ERR, "chdir failed. Error = %m");
          exit (EXIT_FAILURE);
        }
      /* Redirect std*, pidfile FD will be 3, avoids confusion  */
      fclose (stdin);
      fclose (stdout);
      fclose (stderr);
      stdin = fopen ("/dev/null", "r");
      stdout = fopen ("/dev/null", "w");
      stderr = fopen ("/dev/null", "w");
      /* The file lock would not survive a fork, so now: */
    }
  pid_file ();
  /* write values for infodell to pidfile, don't exit on failure */
  data_write ();
  /* Ensure temp2 is the higher value */
  if (dat.t01 > dat.t02)
    {
      tmp = dat.t01;
      dat.t01 = dat.t02;
      dat.t02 = tmp;
    }
  if (dat.t11 > dat.t12)
    {
      tmp = dat.t11;
      dat.t11 = dat.t12;
      dat.t12 = tmp;
    }
  /* Create a 5 C gap between the temperature thresholds if needed */
  if ((dat.t02 - dat.t01) < 5) dat.t01 = (dat.t02 - 5);
  if ((dat.t12 - dat.t11) < 5) dat.t11 = (dat.t12 - 5);
#ifdef DEBUG
  syslog (LOG_ERR, "About to start daemon loop");
#endif
  /* Start the daemon loop */
  while (1)
    {
      tmp = cpu_t; /* Save old temperature in case of strange errors */
      cpu_t = get_cpu_temp ();
      if ((cpu_t == -EINVAL) || (cpu_t == -EAGAIN)) cpu_t = tmp;
#ifdef DEBUG
        syslog (LOG_ERR, "t01=%d, t02=%d, h01=%d, h02=%d, p=%d, cpu=%d",
                dat.t01, dat.t02, dat.h01, dat.h01, dat.poll, cpu_t);
        syslog (LOG_ERR, "t11=%d, t12=%d, h11=%d, h12=%d, what_fan=%d",
                dat.t11, dat.t12, dat.h11, dat.h12, dat.what_fan);
#endif
      /* Check the current fan speed setting */
      if (dat.what_fan < 2) f_stat0 = get_fan_status (dat.what_fan);
      if (dat.what_fan == 2) /* Determine fan 1 setting for 2 fans */
        {
          f_stat0 = get_fan_status (0); /* Error only loses hysteresis */
          f_stat1 = get_fan_status (1);
          tmp = f_stat1; /* As below for f_stat0 */
          if (((cpu_t < dat.t11) && (f_stat1 == 0)) ||
              (cpu_t < (dat.t11 - dat.h11)))
            f_stat1 = 0;
          else f_stat1 = 1;
          if (((cpu_t >= (dat.t12 - dat.h12)) && (tmp == 2)) ||
              (cpu_t > dat.t12))
            f_stat1 = 2;
        } /* Now find a setting for fan 0 (0 or 1 for a single fan) */
      tmp = f_stat0; /* tmp needed for old f_stat0 value */
      /* What should the fan(s) be set to? */
      if (((cpu_t < dat.t01) && (f_stat0 == 0)) ||
          (cpu_t < (dat.t01 - dat.h01)))
        f_stat0 = 0;
      else f_stat0 = 1; /* If it isn't 0 then it's 1 */
      if (((cpu_t >= (dat.t02 - dat.h02)) && (tmp == 2)) ||
          (cpu_t > dat.t02))
        f_stat0 = 2; /* But 1 may get changed to 2 here */
      /* Set the fan(s) */
      if (dat.what_fan < 2)
        {
#ifdef DEBUG
          syslog (LOG_ERR, "Setting fan %d to %d", dat.what_fan, f_stat0);
          if (set_fan (dat.what_fan, f_stat0))
            syslog (LOG_ERR, "Error from set_fan");
#else
          set_fan (dat.what_fan, f_stat0);
#endif
        }
      else
        {
#ifdef DEBUG
          syslog (LOG_ERR, "Setting fan0: %d, fan1: %d", f_stat0, f_stat1);
          if (set_fan (0, f_stat0))
              syslog (LOG_ERR, "Error from set_fan 0");
          if (set_fan (1, f_stat1))
              syslog (LOG_ERR, "Error from set_fan 1");
#else
          set_fan (0, f_stat0); /* An error would mean the fan wasn't set */
          set_fan (1, f_stat1); /* But it'll get tried again next time round */
#endif
        }
      sleep (dat.poll);
      /* 'sleep' will be interrupted by any signals arriving */
      if ((usr1 > 0) || (fin > 0)) break;
    }
  /* Daemon loop ended, clean up */
  /* If a new fanxdell daemon opens syslog the old daemon's messages might */
  /* not be shown in Debian 7, using rsyslog. */
  set_lock (2);
  errno = 0;
  if (X_FINITE_FAILURE_RETRY (close (fd)))
    {
      syslog (LOG_ERR, "close failed in cleanup, error = %m");
    }
  /* When a new daemon has a descriptor, unlink makes a ghost file, so: */
  if (usr1 == 0)
    {
      errno = 0;
      if (unlink (pid_dir))
        {
          syslog (LOG_ERR, "unlink failed in cleanup, error = %m");
        }
    }
  /* Signal replacement daemon to proceed */
  if (usr1 > 0)
    {
      errno = 0;
      if (kill (newpid, SIGUSR2))
        {
          syslog (LOG_ERR, "kill failed to send SIGUSR2, error = %m");
        }
    }
  /* Unblock then raise signal for correct exit code */
  signal_unblock ();
  raise (sigsave);
  syslog (LOG_ERR, "Fatal signal not fatal - terminating.");
  exit (EXIT_FAILURE);
}

void
typo (void)
{
  fprintf (stderr, usage);
  exit (EXIT_FAILURE);
}

int
get_args (char *in_str)
{
  int count = 0;
  char *input_copy;
  char *hashptr;
  char *tmpbuf;
  size_t len = 0;
  int j, i = 0;

  len = strlen (in_str);
  input_copy = (char*) xmalloc (len + 2);
  strcpy (input_copy, in_str);
  if (!(isspace (input_copy[len - 1])))
    { /* ensure final null terminating loop runs by adding /n if needed */
      input_copy[len] = '\n';
      input_copy[len + 1] = '\0';
#ifdef DEBUG
      printf ("len = %d - no space at end of line fixed\n", len);
#endif
    }
  hashptr = strchr (input_copy, '#');
  if (hashptr)
    {
      hashptr[0] = '\n';
      hashptr[1] = '\0';
    }
  len = strlen (input_copy);
  tmpbuf = (char *) xmalloc (len + 1);
  while (i < (int) len)
    {/* last loop reads \n (isspace) so it null terminates array */
      j = 0;
      argB = (char **) xrealloc (argB, ((count + 1) * (sizeof (char *))));
      argB[count] = NULL;
      if (!count)
        {
          argB[count] = (char *) xmalloc (9 * (sizeof (char *)));
          strcpy (argB[count], "fanxdell");
#ifdef DEBUG
          printf ("argB[%d] = %s\n", count, argB[count]);
#endif
          count++;
          continue; /* argv[0] is fanxdell, need matching array numbers */
        }
      while ((isspace (input_copy[i])) && (i < (int) len))
        { /* whitespace, move on */
          i++;
#ifdef DEBUG
          fprintf (stderr, "isspace : input_copy[%d] = %c len = %d\n",
                   i, input_copy[i], len);
#endif
        }
      while ((!(isspace (input_copy[i]))) && (i < (int) len))
        { /* non-whitespace, copy to buffer */
          tmpbuf[j] = input_copy[i];
#ifdef DEBUG
          fprintf (stderr, "tmpbuf[%d] = %c input_copy[%d] = %c len = %d\n",
                  j, tmpbuf[j], i, input_copy[i], len);
#endif
          j++;
          i++;
        }
      if (j)
        { /* transfer buffer contents to array string element */
          tmpbuf[j] = '\0';
          argB[count] = (char *) xmalloc (j + 1);
          strcpy (argB[count], tmpbuf);
#ifdef DEBUG
          fprintf (stderr, "argB[%d] = %s\n", count, argB[count]);
#endif
        }
      count++;
    }
  count--;
  free (tmpbuf);
  /* return count if valid arguments were found */
  if ((count >= 2) && ((!(strcmp (argB[1], "--auto")))
      || (!(strcmp (argB[1], "-a"))))) return (count);
  if ((count >= 3) && ((!(strcmp (argB[1], "--daemon")))
      || (!(strcmp (argB[1], "-d")))))
    {
      if (!(((strtol (argB[2], NULL, 0)) >2) ||
          ((strtol (argB[2], NULL, 0)) < 0))) return (count);
    }
  /* no arguments found, free vector & return 0 */
  for (count = 0; argB[count] != NULL; count++)
    {
      free (argB[count]);
    }
  free (argB);
  argB = NULL;
  return (0);
}

void
get_ioperm (void)
{
#ifndef UNSAFE_MODE
  int signature = 1;
  unsigned int sgv1 = 0, sgv2 = 0;
#endif
  errno = 0;
  if (ioperm (0x84, 1, 1))
    {
      perror ("Initialisation failure in main");
      if (errno == EPERM)
        fprintf (stderr, "Command must be run as root\n");
      exit (EXIT_FAILURE);
    }
  errno = 0;
  if (ioperm (0xb2, 1, 1))
    {
      perror ("Initialisation failure in main");
      if (errno == EPERM)
        fprintf (stderr, "Command must be run as root\n");
      exit (EXIT_FAILURE);
    }
#ifndef UNSAFE_MODE
  /* Test if this is a compatible Dell BIOS */
  if (!(get_dell_sig (0xfea3, &sgv1, &sgv2)))
    {
      if ((((sgv1 >> 24) & 0xff) == 'D') &&
          (((sgv1 >> 16) & 0xff) == 'I') &&
          (((sgv1 >> 8) & 0xff) == 'A') &&
          ((sgv1 & 0xff) == 'G') &&
          (((sgv2 >> 24) & 0xff) == 'D') &&
          (((sgv2 >> 16) & 0xff) == 'E') &&
          (((sgv2 >> 8) & 0xff) == 'L') &&
          ((sgv2 & 0xff) == 'L'))
        signature++;
    }
  if (!(get_dell_sig (0xffa3, &sgv1, &sgv2)))
    {
      if ((((sgv1 >> 24) & 0xff) == 'D') &&
          (((sgv1 >> 16) & 0xff) == 'I') &&
          (((sgv1 >> 8) & 0xff) == 'A') &&
          ((sgv1 & 0xff) == 'G') &&
          (((sgv2 >> 24) & 0xff) == 'D') &&
          (((sgv2 >> 16) & 0xff) == 'E') &&
          (((sgv2 >> 8) & 0xff) == 'L') &&
          ((sgv2 & 0xff) == 'L'))
        signature++;
    }
  if (!signature)
    {
      fprintf (stderr, "Failed to find DIAG DELL signature in BIOS.\n"
               "This may not be a compatible model, try further research\n");
      exit (EXIT_FAILURE);
    }
#endif
}

/* Main function, parse args and take appropriate actions. */
int
main (int argc, char **argv)
{
  int val1 = 0, val2 = 0, tmp1 = 0, tmp2 = 0;
  char *str1 = NULL;
  ssize_t get_ret = (ssize_t) 0;
  size_t n_bytes = (size_t) 0;
  FILE *stream1;
  int args_read = 0, fg = 0, argA = 0;
  struct stat file_stat;

#ifdef DEBUG
  fprintf (stderr, "Compiled for debug mode, writing to stderr / syslog\n");
#endif
  /* --help option (argp doesn't work in installation environment) */
  if ((argc >= 2) && ((!(strcmp (argv[1], "--help")))
      || (!(strcmp (argv[1], "-?")))))
    {
      printf(usage);
      exit (EXIT_SUCCESS);
    }
  /* --version option */
  if ((argc >= 2) && ((!(strcmp (argv[1], "--version")))
      || (!(strcmp (argv[1], "-V")))))
    {
      printf ("fanxdell version number = %d\n", PROGRAM_VERSION);
      exit (EXIT_SUCCESS);
    }
#ifndef UNSAFE_MODE
  dmi_check ();
#endif
  /* --readargs option with possible --foreground flag */
  if ((argc >= 2) && (!(strcmp (argv[1], "--readargs"))))
    {
      if ((argc >= 3) && (!(strcmp (argv[2], "--foreground")))) fg = 1;
      if (!(stat (USER_ARG_FILE, &file_stat)) &&
          (S_ISREG (file_stat.st_mode)))
        { /* file exists and is a regular file */
          errno = 0;
          stream1 = fopen (USER_ARG_FILE, "r");
          if (stream1 == NULL)
            {
              perror ("fopen");
              fprintf (stderr, "Could not open %s\n", USER_ARG_FILE);
            }
          else
            {
              do
                {
                  get_ret = getline (&str1, &n_bytes, stream1);
                  args_read = get_args (str1);
#ifdef DEBUG
                  fprintf (stderr, "str1 = %s args_read = %d\n",
                           str1, args_read);
#endif
                  if (args_read)
                    {
                      argA = args_read;
                      break;
                    }
                }
              while (get_ret != -1);
              if (str1 != NULL) free (str1);
              fclose (stream1);
            }
        }
      if (!args_read)
        { /* default to -a if --readargs was called but no args were read */
          argA = 2;
          argB = (char **) xmalloc (3 * (sizeof (char *)));
          argB[0] = xmalloc (9 * (sizeof (char *)));
          argB[1] = xmalloc (3 * (sizeof (char *)));
          strcpy (argB[0], "fanxdell");
          strcpy (argB[1], "-a");
          argB[2] = NULL;
        }
    }
  else
    {
      argA = argc;
      argB = argv;
    }
  /* --daemon option arg parsing and validity checks */
  if ((argA >= 3) && ((!(strcmp (argB[1], "--daemon")))
      || (!(strcmp (argB[1], "-d")))))
    {
      get_ioperm ();
      dat.what_fan = strtol (argB[2], NULL, 0);
      if ((dat.what_fan >2) || (dat.what_fan < 0)) typo ();
      /* Exits if fan arg invalid, w/out auto what else to do - guess? */
      if (argA >= 4)
        {
          tmp1 = dat.poll;
          dat.poll = strtol (argB[3], NULL, 0);
          if ((dat.poll >30) || (dat.poll < 1))
            {
              fprintf (stderr, "poll argument %d invalid - reset to %d\n",
                       dat.poll, tmp1);
              dat.poll = tmp1;
            }
        }
      if (argA >= 5)
        {
          tmp1 = dat.t01;
          dat.t01 = strtol (argB[4], NULL, 0);
          if ((dat.t01 >= 127) || (dat.t01 < 0))
            {
              fprintf (stderr, "t01 argument %d invalid - reset to %d\n",
                       dat.t01, tmp1);
              dat.t01 = tmp1;
            }
        }
      if (argA >= 6)
        {
          tmp1 = dat.t02;
          dat.t02 = strtol (argB[5], NULL, 0);
          if ((dat.t02 >= 127) || (dat.t02 < 0))
            {
              fprintf (stderr, "t02 argument %d invalid - reset to %d\n",
                       dat.t02, tmp1);
              dat.t02 = tmp1;
            }
        }
      if (argA >= 7)
        {
          tmp1 = dat.h01;
          dat.h01 = strtol (argB[6], NULL, 0);
          if ((dat.h01 >= 127) || (dat.h01 < 0))
            {
              fprintf (stderr, "h01 argument %d invalid - reset to %d\n",
                       dat.h01, tmp1);
              dat.h01 = tmp1;
            }
        }
      if (argA >= 8)
        {
          tmp1 = dat.h02;
          dat.h02 = strtol (argB[7], NULL, 0);
          if ((dat.h02 >= 127) || (dat.h02 < 0))
            {
              fprintf (stderr, "h02 argument %d invalid - reset to %d\n",
                       dat.h02, tmp1);
              dat.h02 = tmp1;
            }
        }
      if (argA >= 9)
        {
          tmp1 = dat.t11;
          dat.t11 = strtol (argB[8], NULL, 0);
          if ((dat.t11 >= 127) || (dat.t11 < 0))
            {
              fprintf (stderr, "t11 argument %d invalid - reset to %d\n",
                       dat.t11, tmp1);
              dat.t11 = tmp1;
            }
        }
      if (argA >= 10)
        {
          tmp1 = dat.t12;
          dat.t12 = strtol (argB[9], NULL, 0);
          if ((dat.t12 >= 127) || (dat.t12 < 0))
            {
              fprintf (stderr, "t12 argument %d invalid - reset to %d\n",
                       dat.t12, tmp1);
              dat.t12 = tmp1;
            }
        }
      if (argA >= 11)
        {
          tmp1 = dat.h11;
          dat.h11 = strtol (argB[10], NULL, 0);
          if ((dat.h11 >= 127) || (dat.h11 < 0))
            {
              fprintf (stderr, "h11 argument %d invalid - reset to %d\n",
                       dat.h11, tmp1);
              dat.h11 = tmp1;
            }
        }
      if (argA >= 12)
        {
          tmp1 = dat.h12;
          dat.h12 = strtol (argB[11], NULL, 0);
          if ((dat.h12 >= 127) || (dat.h12 < 0))
            {
              fprintf (stderr, "h12 argument %d invalid - reset to %d\n",
                       dat.h12, tmp1);
              dat.h12 = tmp1;
            }
        }
      /* Args look valid, go and do daemon things */
      if (fg)
        {
          fan_daemon (0);
        }
      else
        {
          fan_daemon (1);
        }
    }
  /* --auto option, like --daemon but auto detects fan(s) */
  if ((argA >=2) && ((!(strcmp (argB[1], "--auto")))
      || (!(strcmp (argB[1], "-a")))))
    {
      get_ioperm ();
      for (tmp1 = 0; tmp1 < 1000; tmp1++)
        {
          val1 = get_fan_status (0);
          val2 = get_fan_status (1);
          if ((val1 != -EAGAIN) && (val2 != -EAGAIN)) break;
        }
      if (((val1 == -EAGAIN) || (val2 == -EAGAIN)) || ((val1 == -EINVAL)
          && (val2 == -EINVAL)))
        {
          fprintf (stderr, "fanxdell - WARNING - fans not detected\n");
          fprintf (stderr, "Try -d option for manual fan selection\n");
          fprintf (stderr, "BIOS must support ACPI. fanxdell terminating\n");
          exit (EXIT_FAILURE);
        } /* stderr usually goes into syslog at boot (sysvinit & ?) */
      if ((val1 != -EINVAL) && (val2 != -EINVAL))
        {
          dat.what_fan = 2;
        }
      else
        {
          if (val1 != -EINVAL) dat.what_fan = 0;
          if (val2 != -EINVAL) dat.what_fan = 1;
        }
      if (argA >= 3)
        {
          tmp1 = dat.poll;
          dat.poll = strtol (argB[2], NULL, 0);
          if ((dat.poll > 30) || (dat.poll < 1))
            {
              fprintf (stderr, "Poll argument %d invalid - reset to %d\n",
                       dat.poll, tmp1);
              dat.poll = tmp1;
            }
        }
      if (argA >= 4)
        {
          tmp1 = dat.t01;
          dat.t01 = strtol (argB[3], NULL, 0);
          if ((dat.t01 >= 127) || (dat.t01 < 0))
            {
              fprintf (stderr, "t01 argument %d invalid - reset to %d\n",
                       dat.t01, tmp1);
              dat.t01 = tmp1;
            }
        }
      if (argA >= 5)
        {
          tmp1 = dat.t02;
          dat.t02 = strtol (argB[4], NULL, 0);
          if ((dat.t02 >= 127) || (dat.t02 < 0))
            {
              fprintf (stderr, "t02 argument %d invalid - reset to %d\n",
                       dat.t02, tmp1);
              dat.t02 = tmp1;
            }
        }
      if (argA >= 6)
        {
          tmp1 = dat.h01;
          dat.h01 = strtol (argB[5], NULL, 0);
          if ((dat.h01 >= 127) || (dat.h01 < 0))
            {
              fprintf (stderr, "h01 argument %d invalid - reset to %d\n",
                       dat.h01, tmp1);
              dat.h01 = tmp1;
            }
        }
      if (argA >= 7)
        {
          tmp1 = dat.h02;
          dat.h02 = strtol (argB[6], NULL, 0);
          if ((dat.h02 >= 127) || (dat.h02 < 0))
            {
              fprintf (stderr, "h02 argument %d invalid - reset to %d\n",
                       dat.h02, tmp1);
              dat.h02 = tmp1;
            }
        }
      if (argA >= 8)
        {
          tmp1 = dat.t11;
          dat.t11 = strtol (argB[7], NULL, 0);
          if ((dat.t11 >= 127) || (dat.t11 < 0))
                      {
              fprintf (stderr, "t11 argument %d invalid - reset to %d\n",
                       dat.t11, tmp1);
              dat.t11 = tmp1;
            }

        }
      if (argA >= 9)
        {
          tmp1 = dat.t12;
          dat.t12 = strtol (argB[8], NULL, 0);
          if ((dat.t12 >= 127) || (dat.t12 < 0))
            {
              fprintf (stderr, "t12 argument %d invalid - reset to %d\n",
                       dat.t12, tmp1);
              dat.t12 = tmp1;
            }
        }
      if (argA >= 10)
        {
          tmp1 = dat.h11;
          dat.h11 = strtol (argB[9], NULL, 0);
          if ((dat.h11 >= 127) || (dat.h11 < 0))
            {
              fprintf (stderr, "h11 argument %d invalid - reset to %d\n",
                       dat.h11, tmp1);
              dat.h11 = tmp1;
            }
        }
      if (argA >= 11)
        {
          tmp1 = dat.h12;
          dat.h12 = strtol (argB[10], NULL, 0);
          if ((dat.h12 >= 127) || (dat.h12 < 0))
            {
              fprintf (stderr, "h12 argument %d invalid - reset to %d\n",
                       dat.h12, tmp1);
              dat.h12 = tmp1;
            }
        }
      /* Args look valid, go and do daemon things */
      if (fg)
        {
          fan_daemon (0);
        }
      else
        {
          fan_daemon (1);
        }
    }
  /* --fan option to set fan speed(s) */
  if ((argA >= 3) && ((!(strcmp (argB[1], "--fan")))
      || (!(strcmp (argB[1], "-f")))))
    {
      get_ioperm ();
      /* set tmp1 */
      if (!(strcmp (argB[2], "-"))) tmp1 = -1;
      else tmp1 = strtol (argB[2], NULL, 0);
      /* set tmp2 (if there is one) */
      if (argA >= 4)
        {
          if (!(strcmp (argB[3], "-"))) tmp2 = -1;
          else tmp2 = strtol (argB[3], NULL, 0);
        }
      else
        {
          tmp2 = -1;
        }
      if ((tmp1 > 2) || (tmp1 < -1) || (tmp2 > 2) || (tmp2 < -1)) typo ();
      /* Set the fan(s) */
      if (tmp1 >= 0) val1 = set_fan(0, tmp1);
      if (val1 == -EINVAL) fprintf (stderr, "Unable to set fan 0\n");
      if (tmp2 >= 0) val2 = set_fan(1, tmp2);
      if (val2 == -EINVAL) fprintf (stderr, "Unable to set fan 1\n");
#ifdef DEBUG
      fprintf(stderr, "tmp1 = %d, tmp2 = %d\n", tmp1, tmp2);
      fprintf(stderr, "set_fan returned: val1 = %d, val2 = %d\n", val1, val2);
#endif
      exit (EXIT_SUCCESS);
    }
  /* --kill option */
  if ((argA >= 2) && ((!(strcmp (argB[1], "--kill")))
      || (!(strcmp (argB[1], "-k")))))
    {
      get_ioperm ();
      kill_option ();
    }
  /* End of options */
  fprintf (stderr,"-------???????  Could not parse a recognised option\n");
  typo ();
  return (EXIT_FAILURE);
}
