/*
 * infodell, a helper program for fanxdell. Displays BIOS and fan info.
 * Copyright © 2015 - 2016 C_Xperimental
 * Not associated in any way with the company Dell.
 * Not associated in any way with i8kutils or dellfand, although several
 * functions have been adapted from their source code. The interface with
 * the BIOS is very similar but the rest is original.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * <www.cxperimental.weebly.com>
 *
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/io.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <fcntl.h>

/* Enable / disable debugging output */
/*#define DEBUG 1*/

/* Uncomment line below to disable make/model safety checks, then compile. */
/*#define UNSAFE_MODE 1*/

#define PID_FILE_NAME "fanxdell_8sl7m21wdym4.pid" /* Modifiable, malloc'ed */
#define DMI_MAKE "/sys/class/dmi/id/sys_vendor"
#define DMI_MODEL "/sys/class/dmi/id/product_name"
#define PROGRAM_VERSION 3

/* Signal interrupt handling */
#define X_FINITE_FAILURE_RETRY(expression) \
  (__extension__                                                              \
    ({ long int result;                                                       \
       long int index = 0;                                                    \
       do result = (long int) (expression);                                   \
       while (result == -1L && errno == EINTR && (index++) < 1000);          \
       result; }))

typedef struct
{
  unsigned int eax;
  unsigned int ebx __attribute__ ((packed));
  unsigned int ecx __attribute__ ((packed));
  unsigned int edx __attribute__ ((packed));
  unsigned int esi __attribute__ ((packed));
  unsigned int edi __attribute__ ((packed));
} SMMRegisters;

struct fan_data
{
  int what_fan; /* 0 left, 1 right, 2 both */
  int poll;
  int t01;
  int t02;
  int h01;
  int h02;
  int t11;
  int t12;
  int h11;
  int h12;
};

char *pid_dir;
int fd;
char usage[] =
"infodell  : (without options) prints system info to stdout\n\n\
infodell --help / -?  : prints this help.\n\n\
infodell --version / -V  : prints program version.\n\
\n";

int dell_smm (SMMRegisters *regs);
int get_cpu_temp (void);
int get_fan_status (int fan);
int get_bios_version (void);
int get_fan_speed (int fan);
int get_dell_sig (int cmd, unsigned int *eax, unsigned int *edx);
int get_docking_state (void);
#ifndef UNSAFE_MODE
void dmi_check (void);
#endif
void get_ioperm (void);
void info_option (void) __attribute__ ((noreturn));
int set_file_position (void);
int read_data (void);
int main (int argc, char **argv);

/* Write 1 byte to 0xb2, 1 byte to 0x84, put CPU in system management mode */
/* 0xb2 = Dell command register, 0x84 = ACPI SMB_DATA[0] Data register zero */
int
dell_smm (SMMRegisters *regs)
{
  int rc;
  unsigned int eax = regs->eax;
  asm volatile ("pushl %%eax\n\t" \
      "movl 0(%%eax),%%edx\n\t" \
      "push %%edx\n\t" \
      "movl 4(%%eax),%%ebx\n\t" \
      "movl 8(%%eax),%%ecx\n\t" \
      "movl 12(%%eax),%%edx\n\t" \
      "movl 16(%%eax),%%esi\n\t" \
      "movl 20(%%eax),%%edi\n\t" \
      "popl %%eax\n\t" \
      "out %%al,$0xb2\n\t" \
      "out %%al,$0x84\n\t" \
      "xchgl %%eax,(%%esp)\n\t"
      "movl %%ebx,4(%%eax)\n\t" \
      "movl %%ecx,8(%%eax)\n\t" \
      "movl %%edx,12(%%eax)\n\t" \
      "movl %%esi,16(%%eax)\n\t" \
      "movl %%edi,20(%%eax)\n\t" \
      "popl %%edx\n\t" \
      "movl %%edx,0(%%eax)\n\t" \
      "lahf\n\t" \
      "shrl $8,%%eax\n\t" \
      "andl $1,%%eax\n" \
      : "=a" (rc)
      : "a" (regs)
      : "%ebx", "%ecx", "%edx", "%esi", "%edi", "memory");
  if ((regs->eax & 0xffff) == 0xffff) return (-EINVAL);
  /* The above case MUST be handled separately for auto fan detection */
  if ((rc != 0) || (regs->eax == eax)) return (-EAGAIN);
  return (0);
}  /* i386: EINVAL = 22 EAGAIN = 11 (- = no confusion with valid result) */

/* Read the cpu temperature. */
int
get_cpu_temp (void)
{
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};
  int temp, err;
  /* Set to 0 on initialisation only. */
  static int prev = 0;
  regs.eax = 0x10a3;
  err = dell_smm (&regs);
  if (err) return (err);
  temp = regs.eax & 0xff;
  /* Sometimes the temperature sensor returns 0x99, which is out of range. */
  /* In this case we return (once) the previous cached value. */
  if (temp > 127)
    {
      temp = prev;
      prev = 127;
    }
  else
    {
      prev = temp;
    }
  return (temp);
}

/* Current fan setting, off, low, or high: ie. 0, 1, or 2 */
int
get_fan_status (int fan)
{
  int err;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};
  regs.eax = 0x00a3;
  regs.ebx = fan & 0xff;
  err = dell_smm (&regs);
  if (err) return (err);
  return (regs.eax & 0xff);
}

/* Read the BIOS version */
int
get_bios_version (void)
{
  int err;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};

  regs.eax = 0x00a6;
  err = dell_smm (&regs);
  if (err) return (err);
  return (regs.eax);
}

/* Read the fan speed in RPM */
int
get_fan_speed (int fan)
{
  int err;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};

  regs.eax = 0x02a3;
  regs.ebx = fan & 0xff;
  err = dell_smm (&regs);
  if (err) return (err);
  return ((regs.eax & 0xffff) * 15); /* 15 = fan speed multiplier */
}

/* Read DIAG DELL signatures from BIOS */
int
get_dell_sig (int cmd, unsigned int *eax, unsigned int *edx)
{
  int err;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};

  regs.eax = cmd;
  err = dell_smm (&regs);
  if (err) return (err);
  *eax = regs.eax;
  *edx = regs.edx;
  return (0);
}

/* Check if laptop is docked */
int
get_docking_state (void)
{
  int err;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};

  regs.eax = 0x40a3;
  err = dell_smm (&regs);
  if (err) return (err);
  return (regs.eax & 0xffff);
}

/* Read sysfs kernel files to check this is a compatible machine */
#ifndef UNSAFE_MODE
void
dmi_check (void)
{
  char *str1 = NULL;
  ssize_t get_ret = (ssize_t) 0;
  size_t n_bytes = (size_t) 0;
  FILE *stream1;
  int dell = 0;
  int livp = 0;

  errno = 0;
  stream1 = fopen (DMI_MAKE, "r");
  if (stream1 == NULL)
    {
      perror ("fopen");
      fprintf (stderr, "Could not open %s\n", DMI_MAKE);
      exit (EXIT_FAILURE);
    }
  do
    {
      get_ret = getline (&str1, &n_bytes, stream1);
      /* Check make is Dell, strstr returns null for no match */
      if (strstr (str1, "Dell"))
        {
#ifdef DEBUG
          puts ("Well, it's a Dell");
#endif
          dell = 1;
          break;
        }
    }
  while (get_ret != -1);
  /* Runs through loop twice in practice, not changing str1 2nd time (eof) */
  if (str1 != NULL)
    {
      free (str1);
      str1 = NULL;
    }
  n_bytes = (size_t) 0; /* str1, n_bytes reset for next getline */
  fclose (stream1);
  errno = 0;
  stream1 = fopen (DMI_MODEL, "r");
  if (stream1 == NULL)
    {
      perror ("fopen");
      fprintf (stderr, "Could not open %s\n", DMI_MODEL);
      exit (EXIT_FAILURE);
    }
  do
    {
      get_ret = getline (&str1, &n_bytes, stream1);
      /* Check for supported model name */
      if ((strstr (str1, "Latitude")) || (strstr (str1, "Inspiron"))
          || (strstr (str1, "Precision")) || (strstr (str1, "MM061"))
          || (strstr (str1, "MP061")) || (strstr (str1, "Vostro")))
        {
#ifdef DEBUG
          puts ("It's a Latitude, Inspiron, Vostro or Precision");
#endif
          livp = 1;
          break;
        }
    }
  while (get_ret != -1);
  fclose (stream1);
  if (str1 != NULL) /* beware the unlikely unexpected seg fault */
    {
      free (str1); /* free the bytes */
    }
  if (dell == 0)
    {
      fprintf (stderr, "Cannot find Dell in %s\n", DMI_MAKE);
    }
  if (livp == 0)
    {
      fprintf (stderr,
               "Latitude, Inspiron, Precision, Vostro, MM061, MP061\
                not found in %s\n", DMI_MODEL);
    }
  if ((dell == 0) || (livp == 0))
    {
      fprintf (stderr,
               "Your computer may be incompatible, try further research\n");
      exit (EXIT_FAILURE);
    }
}
#endif

void
get_ioperm (void)
{
#ifndef UNSAFE_MODE
  int signature = 1;
  unsigned int sgv1 = 0, sgv2 = 0;
#endif
  errno = 0;
  if (ioperm (0x84, 1, 1))
    {
      perror ("Initialisation failure in main");
      if (errno == EPERM)
        fprintf (stderr, "Command must be run as root\n");
      exit (EXIT_FAILURE);
    }
  errno = 0;
  if (ioperm (0xb2, 1, 1))
    {
      perror ("Initialisation failure in main");
      if (errno == EPERM)
        fprintf (stderr, "Command must be run as root\n");
      exit (EXIT_FAILURE);
    }
#ifndef UNSAFE_MODE
  /* Test if this is a compatible Dell BIOS */
  if (!(get_dell_sig (0xfea3, &sgv1, &sgv2)))
    {
      if ((((sgv1 >> 24) & 0xff) == 'D') &&
          (((sgv1 >> 16) & 0xff) == 'I') &&
          (((sgv1 >> 8) & 0xff) == 'A') &&
          ((sgv1 & 0xff) == 'G') &&
          (((sgv2 >> 24) & 0xff) == 'D') &&
          (((sgv2 >> 16) & 0xff) == 'E') &&
          (((sgv2 >> 8) & 0xff) == 'L') &&
          ((sgv2 & 0xff) == 'L'))
        signature++;
    }
  if (!(get_dell_sig (0xffa3, &sgv1, &sgv2)))
    {
      if ((((sgv1 >> 24) & 0xff) == 'D') &&
          (((sgv1 >> 16) & 0xff) == 'I') &&
          (((sgv1 >> 8) & 0xff) == 'A') &&
          ((sgv1 & 0xff) == 'G') &&
          (((sgv2 >> 24) & 0xff) == 'D') &&
          (((sgv2 >> 16) & 0xff) == 'E') &&
          (((sgv2 >> 8) & 0xff) == 'L') &&
          ((sgv2 & 0xff) == 'L'))
        signature++;
    }
  if (!signature)
    {
      fprintf (stderr, "Failed to find DIAG DELL signature in BIOS.\n"
               "This may not be a compatible model, try further research\n");
      exit (EXIT_FAILURE);
    }
#endif
}

void
info_option (void)
{
  int val1 = 0, val2 = 0, err = 0;
  unsigned int sgv1, sgv2;
  SMMRegisters regs = {0, 0, 0, 0, 0, 0};
  struct stat test_stat;

  err = get_dell_sig (0xfea3, &sgv1, &sgv2);
  if ((err == -EINVAL) || (err == -EAGAIN))
    fprintf (stderr, "Unable to read Dell sig1\n");
  else
    printf ("SIG1 eax = %c%c%c%c edx = %c%c%c%c\n", ((sgv1 >> 24) & 0xff),
            ((sgv1 >> 16) & 0xff), ((sgv1 >> 8) & 0xff), (sgv1 & 0xff),
            ((sgv2 >> 24) & 0xff), ((sgv2 >> 16) & 0xff),
            ((sgv2 >> 8) & 0xff), (sgv2 & 0xff));
  err = get_dell_sig (0xffa3, &sgv1, &sgv2);
  if ((err == -EINVAL) || (err == -EAGAIN))
    fprintf (stderr, "Unable to read Dell sig2\n");
  else
    printf ("SIG2 eax = %c%c%c%c edx = %c%c%c%c\n", ((sgv1 >> 24) & 0xff),
            ((sgv1 >> 16) & 0xff), ((sgv1 >> 8) & 0xff), (sgv1 & 0xff),
            ((sgv2 >> 24) & 0xff), ((sgv2 >> 16) & 0xff),
            ((sgv2 >> 8) & 0xff), (sgv2 & 0xff));
  val1 = get_bios_version ();
  if (val1 == -EINVAL) fprintf (stderr, "BIOS version not found\n");
  if (val1 == -EAGAIN) fprintf (stderr, "Error reading BIOS version\n");
  if ((val1 != -EINVAL) && (val1 != -EAGAIN))
    printf ("BIOS version = %c%c%c%c\n", ((val1 >> 24) & 0xff),
            ((val1 >> 16) & 0xff), ((val1 >> 8) & 0xff), (val1 & 0xff));
  val1 = get_cpu_temp ();
  if (val1 == -EINVAL) fprintf (stderr, "CPU temperature not found\n");
  if (val1 == -EAGAIN) fprintf (stderr, "Error reading CPU temperature\n");
  if ((val1 != -EINVAL) && (val1 != -EAGAIN))
    printf ("CPU temp = %d\n", val1);
  val1 = get_fan_status (0);
  if (val1 == -EINVAL) fprintf (stderr, "Fan 0 not found\n");
  if (val1 == -EAGAIN) fprintf (stderr, "Error reading Fan 0 status\n");
  if ((val1 != -EINVAL) && (val1 != -EAGAIN))
    printf ("Fan 0 status = %d\n", val1);
  val1 = get_fan_status (1);
  if (val1 == -EINVAL) fprintf (stderr, "Fan 1 not found\n");
  if (val1 == -EAGAIN) fprintf (stderr, "Error reading Fan 1 status\n");
  if ((val1 != -EINVAL) && (val1 != -EAGAIN))
    printf ("Fan 1 status = %d\n", val1);
  val1 = get_fan_speed (0);
  if (val1 == -EINVAL) fprintf (stderr, "Fan 0 RPM not found\n");
  if (val1 == -EAGAIN) fprintf (stderr, "Error reading Fan 0 RPM\n");
  if ((val1 != -EINVAL) && (val1 != -EAGAIN))
    printf ("Fan 0 RPM = %d\n", val1);
  val1 = get_fan_speed (1);
  if (val1 == -EINVAL) printf ("Fan 1 RPM not found\n");
  if (val1 == -EAGAIN) fprintf (stderr, "Error reading Fan 1 RPM\n");
  if ((val1 != -EINVAL) && (val1 != -EAGAIN))
    printf ("Fan 1 RPM = %d\n", val1);
  regs.eax = 0x0069;
  err = dell_smm (&regs);
  if (err == -EINVAL) printf ("Power status not found\n");
  if (err == -EAGAIN) fprintf (stderr, "Error reading power status\n");
  if (!err)
    {
      val1 = (regs.eax & 0xff);
      switch (val1)
        {
        case 5:
          printf ("Power supply = %d (AC mains)\n"
                  "Battery charging current = %d\n", val1, regs.ecx);
          break;
        case 3:
          printf ("Power supply = %d (Battery)\n", val1);
          break;
        case 0:
          printf ("Power supply = %d (AC, no battery detected)\n", val1);
          break;
        default:
          printf ("Power supply = %d\n", val1);
        }
      if (val1 != 0)
        {
          printf ("Battery nominal maximum capacity = %d\n"
                  "Battery current power remaining = %d\n",
                  regs.edi, regs.edx);
        }
      printf ("Voltage = %dmV\n", regs.ebx);
    }
  val1 = get_docking_state ();
  if (val1 == -EINVAL) printf ("Dock status not found\n");
  if (val1 == -EAGAIN) fprintf (stderr, "Error reading dock status\n");
  if ((val1 != -EINVAL) && (val1 != -EAGAIN))
    printf ("Docking state = %d\n", val1);
  /* Find status of fanxdell fan daemon */
  val1 = 0;
  errno = 0;
  pid_dir = (char *) malloc ((strlen (PID_FILE_NAME)) + (size_t) 12);
  if (pid_dir == NULL)
    {
      perror ("malloc failed");
      exit (EXIT_FAILURE);
    }
  if ((!(stat ("/run/local", &test_stat))) &&
      (S_ISDIR (test_stat.st_mode))) /* Is it a directory? */
    {
      strcpy (pid_dir, "/run/local/");
      strcat (pid_dir, PID_FILE_NAME);
      if ((!(stat (pid_dir, &test_stat))) && (S_ISREG (test_stat.st_mode)))
        {
          /* It's a  data file */
          if (!(read_data ())) val1++;
          val2++;
        }
    }
  if ((!(stat ("/tmp/local", &test_stat))) &&
      (S_ISDIR (test_stat.st_mode)))
    {
      strcpy (pid_dir, "/tmp/local/");
      strcat (pid_dir, PID_FILE_NAME);
      if ((!(stat (pid_dir, &test_stat))) && (S_ISREG (test_stat.st_mode)))
        {
          /* It's a datafile */
          if (!(read_data ())) val1++;
          val2++;
        }
    }
  if (!val2) puts ("No fanxdell daemon is currently running");
  if ((!val1) && (val2)) puts ("No fanxdell daemon settings available");
  if (val2 == 2) /* Unlikely, fanxdell checks for this at startup */
    {
      puts ("Error - files found in both /run and /tmp\n"
            "Reset : fanxdell -k then fanxdell -a (or use desired args)");
    }

  exit (EXIT_SUCCESS);
}

int
set_file_position (void)
{
  int tmp = 0;
  off_t f_pos = -2;

  while (tmp < 5000)
    {
      tmp++; /* On last while tmp = 4999, set to 5000 here */
      errno = 0;
      f_pos = lseek (fd, 0, SEEK_SET);
      if ((f_pos == -1) && (tmp == 5000))
        {
          perror ("lseek failure");
          return (1);
        }
      if ((f_pos != 0) && (tmp == 5000))
        {
          fprintf (stderr, "lseek problem, f_pos = %d", (int) f_pos);
          return (1);
        }
#ifdef DEBUG
      if ((tmp < 10) || (tmp > 4990))
        {
          printf ("tmp = %d f_pos = %d\n", tmp, (int) f_pos);
        }
#endif
      if (f_pos == 0) break;
    }
  return (0);
}

int
read_data ()
{
  int tmp = 0;
  int i;
  ssize_t byte_no = 0;
  ssize_t byte_total = 0;
  struct fan_data datastruct;
  struct flock lk_dat;
  char *databuf;

  errno = 0;
  fd = X_FINITE_FAILURE_RETRY (open (pid_dir, O_RDONLY | O_NONBLOCK,
                                S_IWUSR | S_IRUSR));
  if (fd < 0)
    {
      fprintf (stderr, "Could not open %s : %m\n", pid_dir);
      return (1);
    }
#ifdef DEBUG
  if (fd >= 0) printf ("opened file : %s\n", pid_dir);
#endif
  for (i = 3; i >= 1; i--)
    {
      lk_dat.l_type = F_WRLCK;
      lk_dat.l_whence = SEEK_SET;
      lk_dat.l_start = (off_t) 0;
      lk_dat.l_len = (off_t) 0;
      lk_dat.l_pid = (pid_t) 0;
      errno = 0;
      if (fcntl (fd, F_GETLK, (&lk_dat)))
        {
          perror ("fcntl F_GETLK failed");
          fprintf (stderr, "Could not get lock status for %s\n", pid_dir);
          if (i <= 1) return (1); /* continue and return on last loop only */
          else continue;
        }
      if (lk_dat.l_type == F_UNLCK)
        {
          /* No active lock, stale pidfile */
          if (i == 3) printf ("Unlocked file, checking for daemon -\n");
          printf ("Seconds remaining : %d\n", i);
          /* Loop in case of a daemon being started during this check */
          if (i <= 1)
            {
              printf ("Error : No running daemon found in %s\n", pid_dir);
              return (1);
            }
        }
      if ((lk_dat.l_type == F_WRLCK) && (lk_dat.l_pid > 0))
        {
          printf ("A fanxdell daemon is active, datafile : %s\n", pid_dir);
          break;
        }
      sleep (1);
    } /* end of for loop */
  if (!((lk_dat.l_type == F_WRLCK) && (lk_dat.l_pid > 0)))
    {
      fprintf (stderr, "An unknown error occurred checking file lock\n");
      return (1);
    }
  if (set_file_position ()) return (1); /* Set file position to file start */
  /* allocate buffer for read */
  errno = 0;
  databuf = (char *) malloc (sizeof (struct fan_data));
  if (databuf == NULL)
    {
      perror ("Virtual memory exhausted for databuf");
      exit (EXIT_FAILURE);
    }
  while (tmp < 5000)
    {
      tmp++;
      errno = 0;
      byte_no = read (fd, (void *) &databuf[byte_total],
                      (sizeof (struct fan_data) - (size_t) byte_total));
      if ((tmp == 5000) && (byte_no == (ssize_t) -1))
        {
          perror ("read () failure for databuf");
        }
      if (byte_no > (ssize_t) 0) byte_total += byte_no;
#ifdef DEBUG
      if ((tmp < 5) || (tmp > 4995))
        {
          printf ("databuf: byte_no = %d, tmp = %d, fd = %d "
                  "struct fan_data = %d\n",
                  (int) byte_no, tmp, fd, (int) (sizeof (struct fan_data)));
          printf ("byte_total = %d\n", (int) byte_total);
        }
#endif
      if (byte_total == (ssize_t) (sizeof (struct fan_data))) break;
      if (tmp == 4000)
        {
          /* Delay & reset for unlikely case fanxdell just starting */
          printf ("Checking for data ...\n");
          sleep (3); /* fanxdell could delay for 1m but shouldn't normally */
          if (set_file_position ())
            {
              free (databuf);
              return (1);
            }
          byte_no = (ssize_t) 0;
          byte_total = (ssize_t) 0;
          for (i = 0; i < ((int) sizeof (struct fan_data)); i++)
            {
              databuf[i] = 0;
#ifdef DEBUG
              printf ("databuf[%d] = %c ", i, databuf[i]);
              if (i == 39) printf ("\n");
#endif
            }
          }
      if (tmp == 5000)
        {
          if (byte_total == (ssize_t) 0)
            {
              puts ("No data could be read from file");
              free (databuf);
              return (1);
            }
          if (byte_total > (ssize_t) 0)
            {
              puts ("Incomplete / unknown data read from settings file");
              free (databuf);
              return (1);
            }
        }
    }
  memcpy ((void *) &datastruct, (const void *) databuf,
          (sizeof (struct fan_data)));
  puts ("Current settings for active fanxdell daemon :");
  if (datastruct.what_fan == 2)
    {
      puts ("Operating twin fans, fan 0 and fan 1");
    }
  else
    {
      printf ("Operating a single fan, designation number = %d\n",
              datastruct.what_fan);
    }
  printf ("Polling interval = %d seconds\n"
          "Single fan settings (0 or 1) or fan 0 with twin fans\n"
          "Turn on at low speed  : t01 = %d\n"
          "Switch to high speed  : t02 = %d\n"
          "Low speed hysteresis  : h01 = %d\n"
          "High speed hysteresis : h02 = %d\n"
          "Second fan settings, fan 1, only applicable for twin fans\n"
          "Turn on at low speed  : t11 = %d\n"
          "Switch to high speed  : t12 = %d\n"
          "Low speed hysteresis  : h11 = %d\n"
          "High speed hysteresis : h12 = %d\n", datastruct.poll,
          datastruct.t01, datastruct.t02, datastruct.h01, datastruct.h02,
          datastruct.t11, datastruct.t12, datastruct.h11, datastruct.h12);
  /* Print all settings even with 1 fan in case HDD is used in 2 laptops */
  free (databuf);
  return (0);
}

int
main (int argc, char **argv)
{
#ifdef DEBUG
  puts ("Compiled for debug mode");
#endif
  /* --help option (argp doesn't work in installation environment) */
  if ((argc >= 2) && ((!(strcmp (argv[1], "--help")))
      || (!(strcmp (argv[1], "-?")))))
    {
      printf(usage);
      exit (EXIT_SUCCESS);
    }
  /* --version option */
  if ((argc >= 2) && ((!(strcmp (argv[1], "--version")))
      || (!(strcmp (argv[1], "-V")))))
    {
      printf ("infodell - interface for the fanxdell fan control daemon\n"
              "fanxdell version number = %d\n", PROGRAM_VERSION);
      exit (EXIT_SUCCESS);
    }
#ifndef UNSAFE_MODE
  dmi_check ();
#endif
  get_ioperm ();
  info_option ();
  return (0);
}


