/*
 * install_fanxdell
 * Installer program for fanxdell and infodell
 * Copyright © 2015 - 2016 C_Xperimental
 * Not associated in any way with the company Dell, dellfand, or i8kutils
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at
 * your option) any later version.
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * <www.cxperimental.weebly.com>
 */


#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>

/* Uncomment to enable debug messages */
/*#define DEBUG = 1*/

#define PID_FILE_NAME "fanxdell_8sl7m21wdym4.pid"
#define PROGRAM_VERSION 3
#define X_FINITE_FAILURE_RETRY(expression) \
  (__extension__                                                              \
    ({ long int result;                                                       \
       long int index = 0;                                                    \
       do result = (long int) (expression);                                   \
       while (result == -1L && errno == EINTR && (index++) < 1000);          \
       result; }))
#define EXEC_FAILED(pathstr) ({do  {fprintf (stderr, \
"Error : couldn't execute %s\n", pathstr); exit (EXIT_FAILURE);} while (0);})
#define FAIL_MSG(pathstr) \
        ({fprintf (stderr, "Error : couldn't execute %s\n", pathstr);})

void *xmalloc (size_t size);
void *xrealloc (void *ptr, size_t size);
int executioner (char *exec_location, char **str_array);
void typo (void);
int check_arguments (const char *in_str);
void set_temp_vals (void);
void compile_64_bit (void);
int is_file_locked (char *pid_dir);
int is_fanxdell_running (void);
void clearance (int userval);
void install (int userval);
void uninstall (int userval);
void get_path (void);
int add_path (char *file_name, char **file_path);
int main (void);

char **path_arr = NULL; /* Array for paths read from PATH variable */
char *gcc_path = NULL;
char *strip_path = NULL;
char *patch_path = NULL;
char *mkdir_path = NULL;
char *cp_path = NULL;
char *chmod_path = NULL;
char *updatercd_path = NULL;
char *invokercd_path = NULL;
char *rm_path = NULL;
char *systemctl_path = NULL;
char *rc_service_path = NULL;
char *rc_update_path = NULL;
char *ln_path = NULL;
char *sv_path = NULL;

char explanation[] =
"--daemon / -d fan poll t01 t02 h01 h02 t11 t12 h11 h12\n\
fan is the only required argument. Select 0, 1, or for twin fans use 2\n\
t01 = turn on fan 0 at low speed at this temperature\n\
h12 = hysteresis for high speed threshold (t12) for fan 1, and so on\n\
Give preceding args to set later values, unset values stay at defaults\n\
Temperature thresholds and hysteresis values must be between 0 and 127 c\n\
Polling frequency between 1 and 30 seconds\n\
--auto / -a poll t01 t02 h01 h02 t11 t12 h11 h12\n\
Identical to --daemon option but automatically detects fans\n\
Init daemeon default:  fanxdell -a\n\
Equivalent to: fanxdell -a 10 35 62 5 8 47 62 7 8\n";

void *
xmalloc (size_t size)
{
  errno = 0;
  void *value = malloc (size);
  if (value == 0)
    {
      perror ("virtual memory exhausted");
      exit (EXIT_FAILURE);
    }
  return value;
}

void *
xrealloc (void *ptr, size_t size)
{
  errno = 0;
  void *value = realloc (ptr, size);
  if (value == 0)
    {
      perror ("virtual memory exhausted");
      exit (EXIT_FAILURE);
    }
  return value;
}

int
executioner (char *exec_location, char **str_array)
{
  int status = 0;
  pid_t pid;
  int i;

  errno = 0;
  pid = fork ();
  if (pid == 0)
    {
      /* This is the child process, execute the shell command */
      for (i = 0; str_array[i] != NULL; i++)
        {
          if (str_array[i] != NULL)
            {
              printf ("%s ", str_array[i]);
            }
        }
      printf ("\n");
      execv (exec_location, str_array);
      _exit (EXIT_FAILURE);
    }
  else
    {
      if (pid < 0)
      /* The fork failed, report failure */
        {
          perror ("fork in executioner");
          fprintf (stderr, "Command = %s\n", str_array[0]);
          exit (EXIT_FAILURE);
        }
      else
        {
          /*This is the parent process, wait for the child to complete */
          errno = 0;
          if (waitpid (pid, &status, 0) != pid)
            {
              perror ("waitpid in executioner");
              fprintf (stderr, "Command = %s\n", str_array[0]);
              exit (EXIT_FAILURE);
            }
        }
    }
#ifdef DEBUG
  printf ("Command return value = %d\n", status);
#endif
  return (status);
}

void
typo (void)
{
  fprintf (stderr, "Arguments are incorrect or outside permitted values\n");
  fprintf (stderr, "No changes were made, no software was installed or "
           "uninstalled, terminating\n");
  exit (EXIT_FAILURE);
}

int
check_arguments (const char *in_str)
{
  int flag = 0;
  char *token;
  const char dl[] = " 	";
  char *input_copy;
  char *free_ptr;
  char *unit = NULL;
  int count = 0;
  char **argB = NULL;

  input_copy = (char *) xmalloc ((strlen (in_str)) + 1);
  free_ptr = input_copy;
  strcpy (input_copy, in_str);
  do
    {
      token = strtok (input_copy, dl);
      input_copy = NULL; /* can't free using this pointer now */
      argB = (char **) xrealloc (argB, ((count + 1) * (sizeof (char *))));
      if (token != NULL)
        {
          unit = (char *) xmalloc ((strlen (token)) + 1);
          strcpy (unit, token);
          argB[count] = unit;
#ifdef DEBUG
          printf ("argB[%d] = %s\n", count, argB[count]);
#endif
          unit = NULL;
        }
      else
        {
          argB[count] = NULL;
        }
      count++;
    }
  while (token != NULL);
  /* check for valid arguments, otherwise exit */
  if ((count >= 3) && ((!(strcmp (argB[0], "--daemon")))
      || (!(strcmp (argB[0], "-d")))))
    {
      flag = 1;
      if (((strtol (argB[1], NULL, 0)) >2) ||
          ((strtol (argB[1], NULL, 0)) < 0)) typo ();
      if (count >= 4)
        {
          if (((strtol (argB[2], NULL, 0)) >30) ||
              ((strtol (argB[2], NULL, 0)) < 1)) typo ();
        }
      if (count >= 5)
        {
          if (((strtol (argB[3], NULL, 0)) >= 127) ||
              ((strtol (argB[3], NULL, 0)) < 0)) typo ();
        }
      if (count >= 6)
        {
          if (((strtol (argB[4], NULL, 0)) >= 127) ||
              ((strtol (argB[4], NULL, 0)) < 0)) typo ();
        }
      if (count >= 7)
        {
          if (((strtol (argB[5], NULL, 0)) >= 127) ||
              ((strtol (argB[5], NULL, 0)) < 0)) typo ();
        }
      if (count >= 8)
        {
          if (((strtol (argB[6], NULL, 0)) >= 127) ||
              ((strtol (argB[6], NULL, 0)) < 0)) typo ();
        }
      if (count >= 9)
        {
          if (((strtol (argB[7], NULL, 0)) >= 127) ||
              ((strtol (argB[7], NULL, 0)) < 0)) typo ();
        }
      if (count >= 10)
        {
          if (((strtol (argB[8], NULL, 0)) >= 127) ||
              ((strtol (argB[8], NULL, 0)) < 0)) typo ();
        }
      if (count >= 11)
        {
          if (((strtol (argB[9], NULL, 0)) >= 127) ||
              ((strtol (argB[9], NULL, 0)) < 0)) typo ();
        }
      if (count >= 12)
        {
          if (((strtol (argB[10], NULL, 0)) >= 127) ||
              ((strtol (argB[10], NULL, 0)) < 0)) typo ();
        }
      if (count >= 13) typo ();
    }
   if ((count >= 2) && ((!(strcmp (argB[0], "--auto")))
      || (!(strcmp (argB[0], "-a")))))
    {
      flag = 1;
      if (count >= 3)
        {
          if (((strtol (argB[1], NULL, 0)) >30) ||
              ((strtol (argB[1], NULL, 0)) < 1)) typo ();
        }
      if (count >= 4)
        {
          if (((strtol (argB[2], NULL, 0)) >= 127) ||
              ((strtol (argB[2], NULL, 0)) < 0)) typo ();
        }
      if (count >= 5)
        {
          if (((strtol (argB[3], NULL, 0)) >= 127) ||
              ((strtol (argB[3], NULL, 0)) < 0)) typo ();
        }
      if (count >= 6)
        {
          if (((strtol (argB[4], NULL, 0)) >= 127) ||
              ((strtol (argB[4], NULL, 0)) < 0)) typo ();
        }
      if (count >= 7)
        {
          if (((strtol (argB[5], NULL, 0)) >= 127) ||
              ((strtol (argB[5], NULL, 0)) < 0)) typo ();
        }
      if (count >= 8)
        {
          if (((strtol (argB[6], NULL, 0)) >= 127) ||
              ((strtol (argB[6], NULL, 0)) < 0)) typo ();
        }
      if (count >= 9)
        {
          if (((strtol (argB[7], NULL, 0)) >= 127) ||
              ((strtol (argB[7], NULL, 0)) < 0)) typo ();
        }
      if (count >= 10)
        {
          if (((strtol (argB[8], NULL, 0)) >= 127) ||
              ((strtol (argB[8], NULL, 0)) < 0)) typo ();
        }
      if (count >= 11)
        {
          if (((strtol (argB[9], NULL, 0)) >= 127) ||
              ((strtol (argB[9], NULL, 0)) < 0)) typo ();
        }
      if (count >= 12) typo ();
    }
   if ((count >= 2) && ((!(strcmp (argB[0], "--remove")))
      || (!(strcmp (argB[0], "-r")))))
    {
      if (count >= 3) typo ();
      flag = 2;
    }
  free (free_ptr);
  for (count = 0; argB[count] != NULL; count++)
    {
      free (argB[count]);
    }
  free (argB);
  return (flag);
}

void
set_temp_vals (void)
{
  int file_exists = 1;
  int fflag = 0;
  ssize_t bytes_read = (ssize_t) 0;
  size_t n_bytes = (size_t) 0;
  char *input_str = NULL;
  FILE *conf_file;
  struct stat file_stat;

  printf ("\n**********************************************\n%s", explanation);
  puts ("Examples: --auto 10 40 67\n"
        "(auto, single fan, temp thresholds up 5 C, default poll and "
        "hysteresis)\n"
        "-a 12 40 67 6 9 52 67 8 9\n"
        "(auto, twin fan, 12 sec poll, all temp up 5 C & hysteresis + 1 C)\n"
        "Only enter arguments for fanxdell, not the fanxdell command");

  if (!(stat ("/etc/default/fanxdell", &file_stat)) &&
      (S_ISREG (file_stat.st_mode)))
    {
      /* file exists and is a regular file */
      errno = 0;
      conf_file = fopen ("/etc/default/fanxdell", "r");
      if (conf_file == NULL)
        {
          perror ("fopen failed to open conf file");
          file_exists = 0;
        }
    }
  else
    {
      file_exists = 0; /* file didn't exist */
    }
  if (file_exists) /* read line from file */
    {
      file_exists = 0; /* Set if file has any contents */
      while (bytes_read != -1)
        {
          bytes_read = getline (&input_str, &n_bytes, conf_file);
          if (bytes_read > 0)
            {
              printf ("*** Current user settings : %s", input_str);
              file_exists = 1;
            }
        }
      printf("\n");
      /* memory allocated with an empty file but 'if' just in case */
      if (input_str != NULL) free (input_str);
      input_str = NULL;
      fclose (conf_file);
      n_bytes = (size_t) 0;
    }
  if (!file_exists)
    {
      puts ("*** No user values are currently set");
    }
  bytes_read = (ssize_t) 0;
  puts ("Enter an argument string to be written to /etc/default/fanxdell\n"
        "Enter -r or --remove to delete /etc/default/fanxdell, use default "
        "settings\n"
        "Press enter with no characters to skip this, making no changes:");
  fflush (stdin);
  errno = 0;
  bytes_read = getline (&input_str, &n_bytes, stdin);
  if (bytes_read <= 0)
    {
      perror ("getline failure, no value read");
      exit (EXIT_FAILURE);
    }
  /* enter only, \n = 1 byte read, one char would be 2 bytes */
  if (bytes_read > 1)
    {
      input_str[bytes_read - 1] = ' '; /* turn \n into a space for strcmp */
      fflag = check_arguments (input_str);
      if (fflag == 1) /* valid arguments to write */
        {
          /* open file for write, write argument string */
          errno = 0;
          conf_file = fopen ("/etc/default/fanxdell", "w");
          if (conf_file == NULL)
            {
              perror ("fopen failed, cannot write values, terminating");
              exit (EXIT_FAILURE);
            }
          errno = 0;
          if ((fputs (input_str, conf_file)) == EOF)
            {
              perror ("fputs failed to write to conf file, terminating");
              exit (EXIT_FAILURE);
            }
          puts ("Arguments successfully written to configuration file\n"
                "To restart an existing installation try:\n"
                "(sysvinit) # invoke-rc.d fanxdell restart\n"
                "(systemd) # systemctl restart fanxdell\n"
                "(openrc) # rc-service fanxdell restart\n"
                "(runit) # sv restart fanxdell\n"
                "infodell will confirm correct operation");

          fclose (conf_file);
        }
      if (fflag == 2) /* request to delete file */
        {
          if (file_exists)
            {
              errno = 0;
              if (unlink ("/etc/default/fanxdell"))
                {
                  perror ("Error: unlink didn't delete /etc/default/fanxdell");
                  puts ("Program terminating, please delete file manually");
                  exit (EXIT_FAILURE);
                }
              else
                {
                  puts ("Deleted /etc/default/fanxdell\n"
                        "Defaults to --auto at reboot or fanxdell restart:\n"
                        "(sysvinit) # invoke-rc.d fanxdell restart\n"
                        "(systemd) # systemctl restart fanxdell\n"
                        "(openrc) # rc-service fanxdell restart\n"
                        "(runit) # sv restart fanxdell\n"
                        "infodell will confirm correct operation");
                }
            }
          else
            {
              puts ("File did not exist, deletion not required");
            }
        }
      if (fflag == 0) /* invalid input, terminate with error message */
        {
          typo ();
        }
    }
  free (input_str);
}

void
compile_64_bit (void)
{
  int tmp = 0;
  char *a[] = {"cp", "src/fanxdell.c", "x86-64/", NULL};
  char *b[] = {"cp", "src/infodell.c", "x86-64/", NULL};
  char *c[] = {"patch", "x86-64/fanxdell.c", "x86-64/x86_64fanx.diff", NULL};
  char *d[] = {"patch", "x86-64/infodell.c", "x86-64/x86_64info.diff", NULL};
  char *e[] = {"gcc", "-Wall", "-Wextra", "-O2", "-o", "x86-64/fanxdell",
               "x86-64/fanxdell.c", NULL};
  char *f[] = {"gcc", "-Wall", "-Wextra", "-O2", "-o", "x86-64/infodell",
               "x86-64/infodell.c", NULL};
  char *g[] = {"strip", "-s", "x86-64/fanxdell", NULL};
  char *h[] = {"strip", "-s", "x86-64/infodell", NULL};

  if (add_path ("cp", &cp_path))
    {
      fprintf (stderr, "Error: cp not found in PATH\n");
      tmp++;
    }
  if (add_path ("patch", &patch_path))
    {
      fprintf (stderr, "Error: patch not found in PATH\n");
      tmp++;
    }
  if (add_path ("gcc", &gcc_path))
    {
      fprintf (stderr, "Error: gcc not found in PATH\n");
      tmp++;
    }
  if (add_path ("strip", &strip_path))
    {
      fprintf (stderr, "Error: strip (binutils package) not found in PATH\n");
      tmp++;
    }
  if (tmp)
    {
      printf ("Cannot compile without the required software -\n"
              "Please install the missing items.\n");
      exit (EXIT_FAILURE);
    }
  if (executioner (cp_path, a)) EXEC_FAILED (cp_path);
  if (executioner (cp_path, b)) EXEC_FAILED (cp_path);
  if (executioner (patch_path, c)) EXEC_FAILED (patch_path);
  if (executioner (patch_path, d)) EXEC_FAILED (patch_path);
  if (executioner (gcc_path, e)) EXEC_FAILED (gcc_path);
  if (executioner (gcc_path, f)) EXEC_FAILED (gcc_path);
  if (executioner (strip_path, g)) EXEC_FAILED (strip_path);
  if (executioner (strip_path, h)) EXEC_FAILED (strip_path);
}

/* return 1 if there is a daemon, 0 if there isn't  */
int
is_file_locked (char *pid_dir)
{
  int i, fd;
  struct flock lk_dat;

  errno = 0;
  fd = X_FINITE_FAILURE_RETRY (open (pid_dir, O_RDONLY | O_NONBLOCK,
                                S_IWUSR | S_IRUSR));
  if (fd < 0)
    {
      fprintf (stderr, "Could not open %s : %m\n", pid_dir);
      return (0);
    }
#ifdef DEBUG
  if (fd >= 0) printf ("opened file : %s\n", pid_dir);
#endif
  for (i = 3; i >= 1; i--)
    {
      lk_dat.l_type = F_WRLCK;
      lk_dat.l_whence = SEEK_SET;
      lk_dat.l_start = (off_t) 0;
      lk_dat.l_len = (off_t) 0;
      lk_dat.l_pid = (pid_t) 0;
#ifdef DEBUG
      printf ("is_file_locked fd = %d i = %d\n", fd, i);
#endif
      errno = 0;
      if (fcntl (fd, F_GETLK, (&lk_dat)))
        {
          perror ("fcntl F_GETLK failed");
          fprintf (stderr, "Could not get lock status for %s\n", pid_dir);
          if (i <= 1) exit (EXIT_FAILURE);
          else continue;
        }
      if (lk_dat.l_type == F_UNLCK)
        {
          /* No active lock, stale pidfile */
          if (i == 3) printf ("Unlocked file, checking for daemon -\n");
          printf ("Seconds remaining : %d\n", i);
          /* Loop in case of a daemon being started during this check */
          if (i <= 1)
            {
              printf ("No running daemon found in %s\n", pid_dir);
              return (0);
            }
        }
      if ((lk_dat.l_type == F_WRLCK) && (lk_dat.l_pid > 0))
        {
          printf ("A fanxdell daemon is active, datafile : %s\n", pid_dir);
          break;
        }
      sleep (1);
    } /* end of for loop */
  if (!((lk_dat.l_type == F_WRLCK) && (lk_dat.l_pid > 0)))
    {
      fprintf (stderr, "An unknown error occurred checking file lock\n");
      return (0);
    }
  return (1);
}

/* like above, return 1 if there is a daemon, 0 if there isn't  */
int
is_fanxdell_running (void)
{
  char *pid_dir;
  struct stat test_stat;

  pid_dir = (char *) xmalloc ((strlen (PID_FILE_NAME)) + (size_t) 12);
  if ((!(stat ("/run/local", &test_stat))) &&
      (S_ISDIR (test_stat.st_mode))) /* Is it a directory? */
    {
      strcpy (pid_dir, "/run/local/");
      strcat (pid_dir, PID_FILE_NAME);
      if ((!(stat (pid_dir, &test_stat))) && (S_ISREG (test_stat.st_mode)))
        {
          /* It's a  data file */
          if (is_file_locked (pid_dir)) return (1);
        }
    }
  if ((!(stat ("/tmp/local", &test_stat))) &&
      (S_ISDIR (test_stat.st_mode)))
    {
      strcpy (pid_dir, "/tmp/local/");
      strcat (pid_dir, PID_FILE_NAME);
      if ((!(stat (pid_dir, &test_stat))) && (S_ISREG (test_stat.st_mode)))
        {
          /* It's a datafile */
          if (is_file_locked (pid_dir)) return (1);
        }
    }
  return 0; /* nothing found */
}

/* Find out if i8kmon is running, also if fanxdell is already installed
 * systemctl status <name> - returns 0 = running, 3 = not, or not found
 * invoke-rc.d <name> status returns 0 if running, 102 else
 * rc-service <name> status returns 0 = running, 3 = not, 1 = not found
 * sv status <name> returns 0 = exists, 1 = not found, n = names not found */

void
clearance (int userval)
{
  int noi8k = 0;
  char *a[] = {"invoke-rc.d",  "i8kmon", "status", NULL};
  char *b[] = {"systemctl", "status", "i8kmon", NULL};
  char *c[] = {"rc-service", "i8kmon", "status", NULL};
  char *d[] = {"sv", "status", "i8kmon", NULL};

  if ((userval == 1) || (userval == 2))
    {/* sysvinit */
      if (add_path ("invoke-rc.d", &invokercd_path))
        {
#ifdef DEBUG
          fprintf (stderr, "Error : invoke-rc.d not found in PATH\n");
#endif
        }
      else
        {
          if (!(executioner (invokercd_path, a)))
            {
              fprintf (stderr, "Error : Detected i8kmon running as daemon\n"
                    "It would conflict with fanxdell - suggested actions:\n"
                    "# invoke-rc.d i8kmon stop\n"
                    "This will stop i8kmon until the next boot, and use:\n"
                    "# update-rc.d i8kmon disable /"
                    " # update-rc.d i8kmon enable\n"
                    "to disable / enable i8kmon for subsequent boots\n");
              exit (EXIT_FAILURE);
            }
          else
            {
              puts ("i8kmon not running, OK to proceed");
              noi8k++;
            }
        }
    }
  if ((userval == 3) || (userval == 4))
    { /* systemd */
      if (add_path ("systemctl", &systemctl_path))
        {
#ifdef DEBUG
          fprintf (stderr, "Error : systemctl not found in PATH\n");
#endif
        }
      else
        {
          if (!(executioner (systemctl_path, b)))
            {
              fprintf (stderr, "Error : Detected i8kmon running as daemon\n"
                   "It would conflict with fanxdell - suggested actions:\n"
                   "# systemctl stop i8kmon\n"
                   "This will stop i8kmon until the next boot, and use:\n"
                   "# systemctl disable i8kmon / # systemctl enable i8kmon\n"
                   "to disable / enable i8kmon for subsequent boots\n");
              exit (EXIT_FAILURE);
            }
          else
            {
              puts ("i8kmon not running, OK to proceed");
              noi8k++;
            }
        }
    }
  if ((userval == 5) || (userval == 6))
    { /* openrc */
      if (add_path ("rc-service", &rc_service_path))
        {
#ifdef DEBUG
          fprintf (stderr, "Error : rc-service not found in PATH\n");
#endif
        }
      else
        {
          if (!(executioner (rc_service_path, c)))
            {
              fprintf (stderr, "Error : Detected i8kmon running as daemon\n"
                   "It would conflict with fanxdell - suggested actions:\n"
                   "# rc-service i8kmon stop\n"
                   "This will stop i8kmon until the next boot, and use:\n"
                   "# rc-update delete i8kmon -a / # rc-update add i8kmon default\n"
                   "to disable / enable i8kmon for subsequent boots\n");
              exit (EXIT_FAILURE);
            }
          else
            {
              puts ("i8kmon not running, OK to proceed");
              noi8k++;
            }
        }
    }
  if ((userval == 7) || (userval == 8))
    { /* runit */
      if (add_path ("sv", &sv_path))
        {
#ifdef DEBUG
          fprintf (stderr, "Error : sv not found in PATH\n");
#endif
        }
      else
        {
          if (!(executioner (sv_path, d)))
            {
              fprintf (stderr, "Error : Detected i8kmon service installed\n"
                   "It would conflict with fanxdell - suggest using:\n"
                   "# sv down i8kmon\n"
                   "# rm /etc/runit/runsvdir/default/i8kmon\n"
                   "# ln -s /etc/sv/i8kmon /etc/runit/runsvdir/default/i8kmon\n"
                   "# sv up i8kmon\n"
                   "to disable / enable i8kmon service\n");
              exit (EXIT_FAILURE);
            }
          else
            {
              puts ("i8kmon not running, OK to proceed");
              noi8k++;
            }
        }
    }
  if (!noi8k)
    {
      puts ("Warning: couldn't determine if i8kmon was running");
    }
  /* Check for running fanxdell (delete binary of active daemon can = probs) */
  if (is_fanxdell_running ())
    {
      puts ("Detected fanxdell running - use : fanxdell -k or"
            " uninstall before installation");
      exit (EXIT_FAILURE);
    }
}

void
install (int userval)
{
  int i, tmp = 0, tmp2 = 0;
  /* actions common to both init systems */
  char *a[] = {"mkdir", "-p", "/usr/local/bin", NULL};
  char *b[] = {"cp", "bin/fanxdell", "/usr/local/bin/", NULL};
  char *c[] = {"cp", "bin/infodell", "/usr/local/bin/", NULL};
  char *d[] = {"cp", "x86-64/fanxdell", "/usr/local/bin/", NULL}; /* 64 */
  char *e[] = {"cp", "x86-64/infodell", "/usr/local/bin/", NULL}; /* 64 */
  char *f[] = {"chmod",  "u+s", "/usr/local/bin/infodell", NULL};
  /* sysvinit */
  char *v1[] = {"mkdir", "-p", "/etc/init.d", NULL};
  char *v2[] = {"cp", "init/fanxdell",  "/etc/init.d/", NULL};
  char *v3[] = {"update-rc.d",  "fanxdell", "defaults", NULL};
  char *v4[] = {"invoke-rc.d",  "fanxdell", "start", NULL};
  char *v5[] = {"invoke-rc.d",  "fanxdell", "status", NULL};
  /* systemd */
  char *d1[] = {"mkdir", "-p", "/usr/local/lib/systemd/system", NULL};
  char *d2[] = {"cp", "init/fanxdell.service",  "/usr/local/lib/systemd/system/", NULL};
  char *d3[] = {"systemctl",  "daemon-reload", NULL};
  char *d4[] = {"systemctl",  "enable", "fanxdell", NULL};
  char *d5[] = {"systemctl",  "start", "fanxdell", NULL};
  char *d6[] = {"systemctl",  "status", "fanxdell", NULL};
  /* openrc */
  char *o1[] = {"mkdir", "-p", "/etc/init.d", NULL};
  char *o2[] = {"cp", "init/openrc/fanxdell",  "/etc/init.d/", NULL};
  char *o3[] = {"chmod",  "a+x", "/etc/init.d/fanxdell", NULL};
  char *o4[] = {"rc-update", "add", "fanxdell", "default", NULL};
  char *o5[] = {"rc-service", "fanxdell", "start", NULL};
  char *o6[] = {"rc-service", "fanxdell", "status", NULL};
  /* runit */
  char *r1[] = {"mkdir", "-p", "/etc/sv/fanxdell", NULL};
  char *r2[] = {"cp", "init/run",  "/etc/sv/fanxdell/", NULL};
  char *r3[] = {"chmod",  "a+x", "/etc/sv/fanxdell/run", NULL};
  char *r4[] = {"ln", "-s", "/run/runit/supervise.fanxdell", "/etc/sv/fanxdell/supervise", NULL};
  char *r5[] = {"ln", "-s", "/etc/sv/fanxdell", "/etc/runit/runsvdir/default/fanxdell", NULL};
  char *r6[] = {"sv",  "up", "fanxdell", NULL};
  /* init system install failed */
  char *z[] = {"fanxdell",  "-a", NULL};

#ifdef DEBUG
  printf ("userval = %d\n", userval);
#endif
  if (add_path ("mkdir", &mkdir_path))
    {
      fprintf (stderr, "Error: mkdir not found in path\n");
      tmp++;
    }
  if ((userval == 1) || (userval == 3) || (userval == 5) || (userval == 7))
    { /* 32 bit only, already done in compile_64_bit function if used */
      if (add_path ("cp", &cp_path))
        {
          fprintf (stderr, "Error: cp not found in path\n");
          tmp++;
        }
    }
  if (add_path ("chmod", &chmod_path))
    {
      fprintf (stderr, "Error: chmod not found in path\n");
      tmp++;
    }
  if (tmp)
    {
      printf ("Cannot install fanxdell without the required software -\n"
              "Please install the missing items.\n");
      exit (EXIT_FAILURE);
    }
  /* set new values for fanxdell if required in /etc/default/fanxdell */
  set_temp_vals ();
  /* execute shell commands */
  if (executioner (mkdir_path, a)) EXEC_FAILED (mkdir_path);
  if ((userval == 1) || (userval == 3) || (userval == 5) || (userval == 7))
    { /* Copy 32 bit versions */
      if (executioner (cp_path, b)) EXEC_FAILED (cp_path);
      if (executioner (cp_path, c)) EXEC_FAILED (cp_path);
    }
  if ((userval == 2) || (userval == 4) || (userval == 6) || (userval == 8))
    { /* Copy 64 bit versions */
      if (executioner (cp_path, d)) EXEC_FAILED (cp_path);
      if (executioner (cp_path, e)) EXEC_FAILED (cp_path);
    }
  if (executioner (chmod_path, f)) EXEC_FAILED (chmod_path);
  tmp = 0;
  if ((userval == 1) || (userval == 2)) /* sysvinit */
    {
      if (add_path ("update-rc.d", &updatercd_path))
        {
          fprintf (stderr, "Error: update-rc.d not found\n");
          tmp2++;
        }
      /* Did add_path () for invoke-rc.d in clearance (), use result */
      if (invokercd_path == NULL) tmp2++;
      if (executioner (mkdir_path, v1)) EXEC_FAILED (mkdir_path);
      if (executioner (cp_path, v2)) EXEC_FAILED (cp_path);
      if (!tmp2)
        {
          if (executioner (updatercd_path, v3)) EXEC_FAILED (updatercd_path);
          if (executioner (invokercd_path, v4)) EXEC_FAILED (invokercd_path);
          if (executioner (invokercd_path, v5))
            {
              fprintf (stderr, "Error : fanxdell not running\n");
            }
          else
            {
#ifdef DEBUG
              printf ("%s : fanxdell installed & running\n", invokercd_path);
#endif
              tmp++;
            }
        }
    }
  if ((userval == 3) || (userval == 4)) /* systemd */
    {
      /* Did add_path () for systemctl in clearance (), use result */
      if (systemctl_path == NULL) tmp2++;
      if (executioner (mkdir_path, d1)) EXEC_FAILED (mkdir_path);
      if (executioner (cp_path, d2)) EXEC_FAILED (cp_path);
      if (!tmp2)
        {
          if (executioner (systemctl_path, d3)) EXEC_FAILED (systemctl_path);
          if (executioner (systemctl_path, d4)) EXEC_FAILED (systemctl_path);
          if (executioner (systemctl_path, d5)) EXEC_FAILED (systemctl_path);
          if (executioner (systemctl_path, d6))
            {
              fprintf (stderr, "Error : fanxdell not running\n");
            }
          else
            {
#ifdef DEBUG
              printf ("%s : fanxdell installed & running\n", invokercd_path);
#endif
              tmp++;
            }
        }
    }
  if ((userval == 5) || (userval == 6)) /* openrc */
    {
      if (add_path ("rc-update", &rc_update_path))
        {
          fprintf (stderr, "Error: rc-update not found\n");
          tmp2++;
        }
      /* Did add_path () for rc-service in clearance (), use result */
      if (rc_service_path == NULL) tmp2++;
      if (executioner (mkdir_path, o1)) EXEC_FAILED (mkdir_path);
      if (executioner (cp_path, o2)) EXEC_FAILED (cp_path);
      if (executioner (chmod_path, o3)) EXEC_FAILED (chmod_path);
      if (!tmp2)
        {
          if (executioner (rc_update_path, o4)) EXEC_FAILED (rc_update_path);
          if (executioner (rc_service_path, o5)) EXEC_FAILED (rc_service_path);
          if (executioner (rc_service_path, o6))
            {
              fprintf (stderr, "Error : fanxdell not running\n");
            }
          else
            {
#ifdef DEBUG
              printf ("%s : fanxdell installed & running\n", rc_service_path);
#endif
              tmp++;
            }
        }
    }
  if ((userval == 7) || (userval == 8)) /* runit */
    {
      if (add_path ("ln", &ln_path))
        {
          fprintf (stderr, "Error: ln not found\n");
          tmp2++;
        }
      /* Did add_path () for sv in clearance (), use result */
      if (sv_path == NULL) tmp2++;
      if (executioner (mkdir_path, r1)) EXEC_FAILED (mkdir_path);
      if (executioner (cp_path, r2)) EXEC_FAILED (cp_path);
      if (executioner (chmod_path, r3)) EXEC_FAILED (chmod_path);
      if (!tmp2)
        {
          if (executioner (ln_path, r4)) EXEC_FAILED (ln_path);
          if (executioner (ln_path, r5)) EXEC_FAILED (ln_path);
          puts ("Waiting for runit to update...");
          for (i = 10 ; i >= 1 ; i--)
            {
              printf ("%d\n", i);
              sleep (1);
            }
          if (executioner (sv_path, r6))
            {
              fprintf (stderr, "Error : fanxdell service not found by runit\n");
            }
          else
            {
#ifdef DEBUG
              printf ("%s : fanxdell installed & is probably running\n", rc_service_path);
#endif
              tmp++;
            }
        }
    }
  if (!tmp) /* tmp should be 1 after any init install completes */
    {
      if (executioner ("/usr/local/bin/fanxdell", z))
          EXEC_FAILED ("/usr/local/bin/fanxdell");
      fprintf (stderr, "Installation incomplete - fanxdell started "
               "but won't start on reboot\n"
               "Error occurred or an unknown init system in use?\n"
               "Please configure your init system to operate fanxdell\n");
    }
}

int
stat_file (char *the_file)
{
  struct stat file_stat;

  errno = 0;
  if ((!(stat (the_file, &file_stat))) &&
      ((S_ISREG (file_stat.st_mode)) || (S_ISLNK (file_stat.st_mode))))
    {
      /* file exists and is a regular file or link */
      return (0);
    }
  else
    {
      fprintf (stderr, "Cannot find %s : %m\n", the_file);
      return (1); /* file didn't exist */
    }
}

/*
 * systemd with sysvinit compat:
 * If fanxdell is installed with a .service file then it can only be
 * controlled with systemctl, not invoke-rc.d.
*/

void
uninstall (int userval)
{
  int tmp = 0;
  int binfan = 0, bininf = 0, deffan = 0, inifan = 0, sysfan = 0;
  /* sysvinit */
  char *v1[] = {"invoke-rc.d", "fanxdell", "stop", NULL};
  char *v2[] = {"rm", "-f", "/etc/init.d/fanxdell", NULL}; /* rm then remove */
  char *v3[] = {"update-rc.d", "fanxdell", "remove", NULL};
  /* systemd */
  char *d1[] = {"systemctl", "disable", "fanxdell", NULL};
  char *d2[] = {"systemctl", "stop", "fanxdell", NULL};
  char *d3[] = {"rm", "-f", "/usr/local/lib/systemd/system/fanxdell.service", NULL};
  char *d4[] = {"systemctl", "daemon-reload", NULL};
  /* openrc */
  char *o1[] = {"rc-service", "fanxdell", "stop", NULL};
  char *o2[] = {"rc-update", "delete", "fanxdell", "default", NULL};
  char *o3[] = {"rm", "-f", "/etc/init.d/fanxdell", NULL};
  /* runit */
  char *r1[] = {"sv", "down","fanxdell", NULL};
  char *r2[] = {"rm", "-f", "/etc/runit/runsvdir/default/fanxdell", NULL};
  char *r3[] = {"rm", "-r", "/etc/sv/fanxdell", NULL};
  /* stop fanxdell if init commands failed to do so */
  char *a[] = {"fanxdell", "-k", NULL};
  /* common actions */
  char *b[] = {"rm", "-f", "/usr/local/bin/fanxdell", NULL};
  char *c[] = {"rm", "-f", "/usr/local/bin/infodell", NULL};
  char *d[] = {"rm", "-f", "/etc/default/fanxdell", NULL};

  if (stat_file ("/usr/local/bin/fanxdell")) binfan++;
  if (stat_file ("/usr/local/bin/infodell")) bininf++;
  if (stat_file ("/etc/default/fanxdell")) deffan++;
  if (stat_file ("/etc/init.d/fanxdell")) inifan++;
  if (stat_file ("/usr/local/lib/systemd/system/fanxdell.service")) sysfan++;
  if (binfan && bininf && deffan && inifan && sysfan)
    {
      puts ("fanxdell is not installed, no action required");
      return;
    }
  if (add_path ("rm", &rm_path))
    {
      fprintf (stderr, "Cannot uninstall fanxdell without rm\n"
              "Please install rm (coreutils package)\n");
      exit (EXIT_FAILURE);
    }
  if (userval == 9) /* sysvinit */
    {
      if (add_path ("update-rc.d", &updatercd_path))
        {
          fprintf (stderr, "Error : update-rc.d not found\n");
          tmp++;
        }
      if (add_path ("invoke-rc.d", &invokercd_path))
        {
          fprintf (stderr, "Error : invoke-rc.d not found\n");
          tmp++;
        }
      if (!tmp)
        {
          if (executioner (invokercd_path, v1)) FAIL_MSG (invokercd_path);
        }
      if (!inifan)
        {
          if (executioner (rm_path, v2)) FAIL_MSG (rm_path);
        }
      if (!tmp)
        {
          if (executioner (updatercd_path, v3)) FAIL_MSG (updatercd_path);
        }
    }
  if (userval == 10) /* systemd */
    {
      if (add_path ("systemctl", &systemctl_path))
        {
          fprintf (stderr, "Error : systemctl not found\n");
          tmp++;
        }
      if (!tmp)
        {
          if (executioner (systemctl_path, d1)) FAIL_MSG (systemctl_path);
          if (executioner (systemctl_path, d2)) FAIL_MSG (systemctl_path);
        }
      if (!sysfan)
        {
          if (executioner (rm_path, d3)) FAIL_MSG (rm_path);
        }
      if (!tmp)
        {
          if (executioner (systemctl_path, d4)) FAIL_MSG (systemctl_path);
        }
    }
  if (userval == 11) /* openrc */
    {
      if (add_path ("rc-update", &rc_update_path))
        {
          fprintf (stderr, "Error : rc-update not found\n");
          tmp++;
        }
      if (add_path ("rc-service", &rc_service_path))
        {
          fprintf (stderr, "Error : rc-service not found\n");
          tmp++;
        }
      if (!tmp)
        {
          if (executioner (rc_service_path, o1)) FAIL_MSG (rc_service_path);
        }
      if (!tmp)
        {
          if (executioner (rc_update_path, o2)) FAIL_MSG (rc_update_path);
        }
      if (!inifan)
        {
          if (executioner (rm_path, o3)) FAIL_MSG (rm_path);
        }
    }
  if (userval == 12) /* runit */
    {
      if (add_path ("sv", &sv_path))
        {
          fprintf (stderr, "Error : sv not found\n");
          tmp++;
        }
      if (!tmp)
        {
          if (executioner (sv_path, r1)) FAIL_MSG (sv_path);
        }
      if (executioner (rm_path, r2)) FAIL_MSG (rm_path);
      if (executioner (rm_path, r3)) FAIL_MSG (rm_path);
      tmp++; /* sv return values only say if a service is not found */
    }
  if (tmp) /* daemon commands not present */
    {
      if (executioner ("/usr/local/bin/fanxdell", a))
        FAIL_MSG ("/usr/local/bin/fanxdell");
    }
  if (!binfan)
    {
      if (executioner (rm_path, b)) FAIL_MSG (rm_path);
    }
  if (!bininf)
    {
      if (executioner (rm_path, c)) FAIL_MSG (rm_path);
    }
  if (!deffan)
    {
      if (executioner (rm_path, d)) FAIL_MSG (rm_path);
    }
}

void
get_path (void)
{
  char *token;
  const char dl[] = ":";
  char *orig_path;
  char *path_copy;
  char *free_ptr;
  char *unit = NULL;
  int count = 0;

  orig_path = (getenv ("PATH"));
  if (orig_path == NULL)
    {
      fprintf (stderr,
               "Unable to read PATH environment variable, terminating\n");
      exit (EXIT_FAILURE);
    }
  printf ("PATH = %s\n", orig_path);
  path_copy = (char *) xmalloc ((strlen (orig_path)) + 1);
  free_ptr = path_copy;
  strcpy (path_copy, orig_path);
/* Write tokens to array with unknown no. of strings of unknown length */
  do
    {
      token = strtok (path_copy, dl);
      path_copy = NULL; /* Set to NULL for strtok after first loop */
      path_arr = (char **) xrealloc (path_arr,
                                    ((count + 1) * (sizeof (char *))));
      if (token != NULL)
        { /* strlen on NULL token gives run time segmentation fault */
#ifdef DEBUG
          printf ("getenv token = %s length = %d\n",
                 token, (int) (strlen (token)));
#endif
          unit = (char *) xmalloc ((strlen (token)) + 1);
          strcpy (unit, token);
          path_arr[count] = unit;
          unit = NULL;
        }
      else
        {
          path_arr[count] = NULL;
        }
      count++;
    }
  while (token != NULL);/* No tokens found case is handled by add_path */
  free (free_ptr);
}

int
add_path (char *file_name, char **file_path)
{
  int i;
  struct stat file_stat;

  for (i = 0; path_arr[i] != NULL; i++) /* examine each possible path */
    {
#ifdef DEBUG
      printf ("path[%d] = %s\n", i, path_arr[i]);
#endif
      *file_path = (char *) xrealloc (*file_path, (strlen (file_name)) +
                                    (strlen (path_arr[i])) + (size_t) 2);
      strcpy (*file_path, path_arr[i]);
      strcat (*file_path, "/"); /* add / on end of path */
      strcat (*file_path, file_name);
      if (!(stat (*file_path, &file_stat)) && (file_stat.st_mode & S_IXUSR)
          && ((S_ISREG (file_stat.st_mode)) || (S_ISLNK (file_stat.st_mode))))
        {
          /* file exists and has execute permission for file owner */
          /* and is a regular file or a link. */
#ifdef DEBUG
          printf ("file_path = %s\n", *file_path);
#endif
          return 0; /* return successfully if file was found */
        }
    }
#ifdef DEBUG
  printf ("Did not find %s in path.\n", file_name);
#endif
  if (*file_path != NULL)
    {
      free (*file_path);
      *file_path = NULL;
    }
  return 1;
}

int
main (void)
{
  int val;
  ssize_t bytes_read = 0;
  size_t n_bytes = (size_t) 0;
  char *input_str = NULL;
  uid_t what_user;

  what_user = getuid ();
  if (what_user)
    {
      printf ("**WARNING** Root permission is required to run this program\n"
              "Detected user ID = %d, usually root ID is 0\n", what_user);
    }
  get_path ();
  printf ("fanxdell version %d - installation program\n", PROGRAM_VERSION);
  puts ("Please type a number to select an option from these choices :\n"
        "1  - Install 32 bit fanxdell and infodell (sysvinit sysvrc)\n"
        "2  - Install 64 bit fanxdell and infodell, experimental (sysvinit sysvrc)\n"
        "3  - Install 32 bit fanxdell and infodell (systemd)\n"
        "4  - Install 64 bit fanxdell and infodell, experimental (systemd)\n"
        "5  - Install 32 bit fanxdell and infodell (openrc)\n"
        "6  - Install 64 bit fanxdell and infodell, experimental (openrc)\n"
        "7  - Install 32 bit fanxdell and infodell (runit)\n"
        "8  - Install 64 bit fanxdell and infodell, experimental (runit)\n"
        "9  - Uninstall fanxdell and infodell (sysvinit sysvrc)\n"
        "10 - Uninstall fanxdell and infodell (systemd)\n"
        "11 - Uninstall fanxdell and infodell (openrc)\n"
        "12 - Uninstall fanxdell and infodell (runit)\n"
        "13 - Set/reset temperature values for previously installed fanxdell\n"
        "14 - Quit, end program and exit now");
  fflush (stdin);
  errno = 0;
  bytes_read = getline (&input_str, &n_bytes, stdin);
  if (bytes_read <= 0)
    {
      perror ("getline failure, no value read");
      exit (EXIT_FAILURE);
    }
  val = strtol (input_str, NULL, 0);
#ifdef DEBUG
  printf ("bytes_read = %d val = %d\n", (int) bytes_read, val);
#endif
  switch (val)
    {
    case 2 :
    case 4 :
    case 6 :
    case 8 :
      compile_64_bit (); /* no break, fall through */
    case 1 :
    case 3 :
    case 5 :
    case 7 :
      clearance (val);
      install (val);
      break;
    case 9  :
    case 10 :
    case 11 :
    case 12 :
      uninstall (val);
      break;
    case 13 :
      set_temp_vals ();
      break;
    case 14 :
      puts ("Terminating now, no software was installed or uninstalled");
      exit (EXIT_SUCCESS);
    default :
      printf ("Invalid number entered : %d\n"
              "Please run the installer again\n", val);
      exit (EXIT_FAILURE);
    }
  puts ("Thankyou for using C_Xperimental software");
  return (0);
}
